"""
Functions for reading NEXRAD level 2 files.
"""



# This file is part of the Py-ART, the Python ARM Radar Toolkit
# https://github.com/ARM-DOE/pyart

# Care has been taken to keep this file free from extraneous dependancies
# so that it can be used by other projects with no/minimal modification.

# Please feel free to use this file in other project provided the license
# below is followed. Keeping the above comment lines would also be helpful
# to direct other back to the Py-ART project and the source of this file.


LICENSE = """
Copyright (c) 2013, UChicago Argonne, LLC
All rights reserved.

Copyright 2013 UChicago Argonne, LLC. This software was produced under U.S.
Government contract DE-AC02-06CH11357 for Argonne National Laboratory (ANL),
which is operated by UChicago Argonne, LLC for the U.S. Department of Energy.
The U.S. Government has rights to use, reproduce, and distribute this
software. NEITHER THE GOVERNMENT NOR UCHICAGO ARGONNE, LLC MAKES ANY
WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS
SOFTWARE. If software is modified to produce derivative works, such modified
software should be clearly marked, so as not to confuse it with the version
available from ANL.

Additionally, redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following conditions
are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of UChicago Argonne, LLC, Argonne National
      Laboratory, ANL, the U.S. Government, nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY UCHICAGO ARGONNE, LLC AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL UCHICAGO ARGONNE, LLC OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import bz2
from datetime import datetime, timedelta
import struct
import warnings

import numpy as np


class NEXRADLevel2File(object):
    """
    Class for accessing data in a NEXRAD (WSR-88D) Level II file.

    NEXRAD Level II files [1]_, also know as NEXRAD Archive Level II or
    WSR-88D Archive level 2, are available from the NOAA National Climate Data
    Center [2]_ as well as on the UCAR THREDDS Data Server [3]_. Files with
    uncompressed messages and compressed messages are supported. This class
    supports reading both "message 31" and "message 1" type files.

    Parameters
    ----------
    filename : str
        Filename of Archive II file to read.

    Attributes
    ----------
    radial_records : list
        Radial (1 or 31) messages in the file.
    nscans : int
        Number of scans in the file.
    scan_msgs : list of arrays
        Each element specifies the indices of the message in the
        radial_records attribute which belong to a given scan.
    volume_header : dict
        Volume header.
    vcp : dict
        VCP information dictionary.
    _records : list
        A list of all records (message) in the file.
    _fh : file-like
        File like object from which data is read.
    _msg_type : '31' or '1':
        Type of radial messages in file.

    References
    ----------
    .. [1] http://www.roc.noaa.gov/WSR88D/Level_II/Level2Info.aspx
    .. [2] http://www.ncdc.noaa.gov/
    .. [3] http://thredds.ucar.edu/thredds/catalog.html

    """

    def __init__(self, filename):
        """ initalize the object. """
        # read in the volume header and compression_record
        if hasattr(filename, 'read'):
            fh = filename
        else:
            fh = open(filename, 'rb')
        size = _structure_size(VOLUME_HEADER)
        self.volume_header = _unpack_structure(fh.read(size), VOLUME_HEADER)
        compression_record = fh.read(COMPRESSION_RECORD_SIZE)

        # read the records in the file, decompressing as needed
        compression_slice = slice(CONTROL_WORD_SIZE, CONTROL_WORD_SIZE + 2)
        compression_or_ctm_info = compression_record[compression_slice]
        if compression_or_ctm_info == b'BZ':
            buf = _decompress_records(fh)
        # The 12-byte compression record previously held the Channel Terminal
        # Manager (CTM) information. Bytes 4 through 6 contain the size of the
        # record (2432) as a big endian unsigned short, which is encoded as
        # b'\t\x80' == struct.pack('>H', 2432).
        # Newer files zero out this section.
        elif compression_or_ctm_info in (b'\x00\x00', b'\t\x80'):
            buf = fh.read()
        else:
            raise IOError('unknown compression record')
        self._fh = fh

        # read the records from the buffer
        self._records = []
        buf_length = len(buf)
        pos = 0
        while pos < buf_length:
            pos, dic = _get_record_from_buf(buf, pos)
            self._records.append(dic)

        # pull out radial records (1 or 31) which contain the moment data.
        self.radial_records = [r for r in self._records
                               if r['header']['type'] == 31]
        self._msg_type = '31'
        if len(self.radial_records) == 0:
            self.radial_records = [r for r in self._records
                                   if r['header']['type'] == 1]
            self._msg_type = '1'
        if len(self.radial_records) == 0:
            raise ValueError('No MSG31 records found, cannot read file')
        elev_nums = np.array([m['msg_header']['elevation_number']
                              for m in self.radial_records])
        self.scan_msgs = [np.where(elev_nums == i + 1)[0]
                          for i in range(elev_nums.max())]
        self.nscans = len(self.scan_msgs)

        # pull out the vcp record
        msg_5 = [r for r in self._records if r['header']['type'] == 5]

        if len(msg_5):
            self.vcp = msg_5[0]
        else:
            # There is no VCP Data.. This is uber dodgy
            warnings.warn("No MSG5 detected. Setting to meaningless data. "
                          "Rethink your life choices and be ready for errors."
                          "Specifically fixed angle data will be missing")

            self.vcp = None
        return

    def close(self):
        """ Close the file. """
        self._fh.close()

    def location(self):
        """
        Find the location of the radar.

        Returns all zeros if location is not available.

        Returns
        -------
        latitude : float
            Latitude of the radar in degrees.
        longitude : float
            Longitude of the radar in degrees.
        height : int
            Height of radar and feedhorn in meters above mean sea level.

        """
        if self._msg_type == '31':
            dic = self.radial_records[0]['VOL']
            height = dic['height'] + dic['feedhorn_height']
            return dic['lat'], dic['lon'], height
        else:
            return 0.0, 0.0, 0.0

    def scan_info(self, scans=None):
        """
        Return a list of dictionaries with scan information.

        Parameters
        ----------
        scans : list ot None
            Scans (0 based) for which ray (radial) azimuth angles will be
            retrieved.  None (the default) will return the angles for all
            scans in the volume.

        Returns
        -------
        scan_info : list, optional
            A list of the scan performed with a dictionary with keys
            'moments', 'ngates', 'nrays', 'first_gate' and 'gate_spacing'
            for each scan.  The 'moments', 'ngates', 'first_gate', and
            'gate_spacing' keys are lists of the NEXRAD moments and gate
            information for that moment collected during the specific scan.
            The 'nrays' key provides the number of radials collected in the
            given scan.

        """
        info = []

        if scans is None:
            scans = range(self.nscans)
        for scan in scans:
            nrays = self.get_nrays(scan)
            if nrays < 2:
                self.nscans -= 1
                continue
            msg31_number = self.scan_msgs[scan][0]
            msg = self.radial_records[msg31_number]
            nexrad_moments = ['REF', 'VEL', 'SW', 'ZDR', 'PHI', 'RHO', 'CFP']
            moments = [f for f in nexrad_moments if f in msg]
            ngates = [msg[f]['ngates'] for f in moments]
            gate_spacing = [msg[f]['gate_spacing'] for f in moments]
            first_gate = [msg[f]['first_gate'] for f in moments]
            info.append({
                'nrays': nrays,
                'ngates': ngates,
                'gate_spacing': gate_spacing,
                'first_gate': first_gate,
                'moments': moments})
        return info

    def get_vcp_pattern(self):
        """
        Return the numerical volume coverage pattern (VCP) or None if unknown.
        """
        if self.vcp is None:
            return None
        else:
            return self.vcp['msg5_header']['pattern_number']

    def get_nrays(self, scan):
        """
        Return the number of rays in a given scan.

        Parameters
        ----------
        scan : int
            Scan of interest (0 based).

        Returns
        -------
        nrays : int
            Number of rays (radials) in the scan.

        """
        return len(self.scan_msgs[scan])

    def get_range(self, scan_num, moment):
        """
        Return an array of gate ranges for a given scan and moment.

        Parameters
        ----------
        scan_num : int
            Scan number (0 based).
        moment : 'REF', 'VEL', 'SW', 'ZDR', 'PHI', 'RHO', or 'CFP'
            Moment of interest.

        Returns
        -------
        range : ndarray
            Range in meters from the antenna to the center of gate (bin).

        """
        dic = self.radial_records[self.scan_msgs[scan_num][0]][moment]
        ngates = dic['ngates']
        first_gate = dic['first_gate']
        gate_spacing = dic['gate_spacing']
        return np.arange(ngates) * gate_spacing + first_gate

    # helper functions for looping over scans
    def _msg_nums(self, scans):
        """ Find the all message number for a list of scans. """
        return np.concatenate([self.scan_msgs[i] for i in scans])

    def _radial_array(self, scans, key):
        """
        Return an array of radial header elements for all rays in scans.
        """
        msg_nums = self._msg_nums(scans)
        temp = [self.radial_records[i]['msg_header'][key] for i in msg_nums]
        return np.array(temp)

    def _radial_sub_array(self, scans, key):
        """
        Return an array of RAD or msg_header elements for all rays in scans.
        """
        msg_nums = self._msg_nums(scans)
        if self._msg_type == '31':
            tmp = [self.radial_records[i]['RAD'][key] for i in msg_nums]
        else:
            tmp = [self.radial_records[i]['msg_header'][key] for i in msg_nums]
        return np.array(tmp)

    def get_times(self, scans=None):
        """
        Retrieve the times at which the rays were collected.

        Parameters
        ----------
        scans : list or None
            Scans (0-based) to retrieve ray (radial) collection times from.
            None (the default) will return the times for all scans in the
            volume.

        Returns
        -------
        time_start : Datetime
            Initial time.
        time : ndarray
            Offset in seconds from the initial time at which the rays
            in the requested scans were collected.

        """
        if scans is None:
            scans = range(self.nscans)
        days = self._radial_array(scans, 'collect_date')
        secs = self._radial_array(scans, 'collect_ms') / 1000.
        offset = timedelta(days=int(days[0]) - 1, seconds=int(secs[0]))
        time_start = datetime(1970, 1, 1) + offset
        time = secs - int(secs[0]) + (days - days[0]) * 86400
        return time_start, time

    def get_azimuth_angles(self, scans=None):
        """
        Retrieve the azimuth angles of all rays in the requested scans.

        Parameters
        ----------
        scans : list ot None
            Scans (0 based) for which ray (radial) azimuth angles will be
            retrieved. None (the default) will return the angles for all
            scans in the volume.

        Returns
        -------
        angles : ndarray
            Azimuth angles in degress for all rays in the requested scans.

        """
        if scans is None:
            scans = range(self.nscans)
        if self._msg_type == '1':
            scale = 180 / (4096 * 8.)
        else:
            scale = 1.
        return self._radial_array(scans, 'azimuth_angle') * scale

    def get_elevation_angles(self, scans=None):
        """
        Retrieve the elevation angles of all rays in the requested scans.

        Parameters
        ----------
        scans : list or None
            Scans (0 based) for which ray (radial) azimuth angles will be
            retrieved. None (the default) will return the angles for
            all scans in the volume.

        Returns
        -------
        angles : ndarray
            Elevation angles in degress for all rays in the requested scans.

        """
        if scans is None:
            scans = range(self.nscans)
        if self._msg_type == '1':
            scale = 180 / (4096 * 8.)
        else:
            scale = 1.
        return self._radial_array(scans, 'elevation_angle') * scale

    def get_target_angles(self, scans=None):
        """
        Retrieve the target elevation angle of the requested scans.

        Parameters
        ----------
        scans : list or None
            Scans (0 based) for which the target elevation angles will be
            retrieved. None (the default) will return the angles for all
            scans in the volume.

        Returns
        -------
        angles : ndarray
            Target elevation angles in degress for the requested scans.

        """
        if scans is None:
            scans = range(self.nscans)
        if self._msg_type == '31':
            if self.vcp is not None:
                cut_parameters = self.vcp['cut_parameters']
            else:
                cut_parameters = [{'elevation_angle': 0.0}] * self.nscans
            scale = 360. / 65536.
            return np.array([cut_parameters[i]['elevation_angle'] * scale
                             for i in scans], dtype='float32')
        else:
            scale = 180 / (4096 * 8.)
            msgs = [self.radial_records[self.scan_msgs[i][0]] for i in scans]
            return np.round(np.array(
                [m['msg_header']['elevation_angle'] * scale for m in msgs],
                dtype='float32'), 1)

    def get_nyquist_vel(self, scans=None):
        """
        Retrieve the Nyquist velocities of the requested scans.

        Parameters
        ----------
        scans : list or None
            Scans (0 based) for which the Nyquist velocities will be
            retrieved. None (the default) will return the velocities for all
            scans in the volume.

        Returns
        -------
        velocities : ndarray
            Nyquist velocities (in m/s) for the requested scans.

        """
        if scans is None:
            scans = range(self.nscans)
        return self._radial_sub_array(scans, 'nyquist_vel') * 0.01

    def get_unambigous_range(self, scans=None):
        """
        Retrieve the unambiguous range of the requested scans.

        Parameters
        ----------
        scans : list or None
            Scans (0 based) for which the unambiguous range will be retrieved.
            None (the default) will return the range for all scans in the
            volume.

        Returns
        -------
        unambiguous_range : ndarray
            Unambiguous range (in meters) for the requested scans.

        """
        if scans is None:
            scans = range(self.nscans)
        # unambiguous range is stored in tenths of km, x100 for meters
        return self._radial_sub_array(scans, 'unambig_range') * 100.

    def get_data(self, moment, max_ngates, scans=None, raw_data=False):
        """
        Retrieve moment data for a given set of scans.

        Masked points indicate that the data was not collected, below
        threshold or is range folded.

        Parameters
        ----------
        moment : 'REF', 'VEL', 'SW', 'ZDR', 'PHI', 'RHO', or 'CFP'
            Moment for which to to retrieve data.
        max_ngates : int
            Maximum number of gates (bins) in any ray.
            requested.
        raw_data : bool
            True to return the raw data, False to perform masking as well as
            applying the appropiate scale and offset to the data.  When
            raw_data is True values of 1 in the data likely indicate that
            the gate was not present in the sweep, in some cases in will
            indicate range folded data.
        scans : list or None.
            Scans to retrieve data from (0 based). None (the default) will
            get the data for all scans in the volume.

        Returns
        -------
        data : ndarray

        """
        if scans is None:
            scans = range(self.nscans)

        # determine the number of rays
        msg_nums = self._msg_nums(scans)
        nrays = len(msg_nums)
        # extract the data
        set_datatype = False
        data = np.ones((nrays, max_ngates), '>B')
        for i, msg_num in enumerate(msg_nums):
            msg = self.radial_records[msg_num]
            if moment not in msg.keys():
                continue
            if not set_datatype:
                data = data.astype('>'+_bits_to_code(msg, moment))
                set_datatype = True

            ngates = min(msg[moment]['ngates'], max_ngates,
                         len(msg[moment]['data']))
            data[i, :ngates] = msg[moment]['data'][:ngates]
        # return raw data if requested
        if raw_data:
            return data

        # mask, scan and offset, assume that the offset and scale
        # are the same in all scans/gates
        for scan in scans:  # find a scan which contains the moment
            msg_num = self.scan_msgs[scan][0]
            msg = self.radial_records[msg_num]
            if moment in msg.keys():
                offset = np.float32(msg[moment]['offset'])
                scale = np.float32(msg[moment]['scale'])
                mask = data <= 1
                scaled_data = (data - offset) / scale
                return np.ma.array(scaled_data, mask=mask)

        # moment is not present in any scan, mask all values
        return np.ma.masked_less_equal(data, 1)


def _bits_to_code(msg, moment):
    """
    Convert number of bits to the proper code for unpacking.
    Based on the code found in MetPy:
    https://github.com/Unidata/MetPy/blob/40d5c12ab341a449c9398508bd41
    d010165f9eeb/src/metpy/io/_tools.py#L313-L321
    """
    if msg['header']['type'] == 1:
        word_size = msg[moment]['data'].dtype
        if word_size == 'uint16':
            return 'H'
        elif word_size == 'uint8':
            return 'B'
        else:
            warnings.warn(
                ('Unsupported bit size: %s. Returning "B"', word_size))
            return 'B'

    elif msg['header']['type'] == 31:
        word_size = msg[moment]['word_size']
        if word_size == 16:
            return 'H'
        elif word_size == 8:
            return 'B'
        else:
            warnings.warn(
                ('Unsupported bit size: %s. Returning "B"', word_size))
            return 'B'
    else:
        raise TypeError("Unsupported msg type %s", msg['header']['type'])


def _decompress_records(file_handler):
    """
    Decompressed the records from an BZ2 compressed Archive 2 file.
    """
    file_handler.seek(0)
    cbuf = file_handler.read()    # read all data from the file
    decompressor = bz2.BZ2Decompressor()
    skip = _structure_size(VOLUME_HEADER) + CONTROL_WORD_SIZE
    buf = decompressor.decompress(cbuf[skip:])
    while len(decompressor.unused_data):
        cbuf = decompressor.unused_data
        decompressor = bz2.BZ2Decompressor()
        buf += decompressor.decompress(cbuf[CONTROL_WORD_SIZE:])

    return buf[COMPRESSION_RECORD_SIZE:]


def _get_record_from_buf(buf, pos):
    """ Retrieve and unpack a NEXRAD record from a buffer. """
    dic = {'header': _unpack_from_buf(buf, pos, MSG_HEADER)}
    msg_type = dic['header']['type']

    if msg_type == 31:
        new_pos = _get_msg31_from_buf(buf, pos, dic)
    elif msg_type == 5:
        # Sometimes we encounter incomplete buffers
        try:
            new_pos = _get_msg5_from_buf(buf, pos, dic)
        except struct.error:
            warnings.warn("Encountered incomplete MSG5. File may be corrupt.",
                          RuntimeWarning)
            new_pos = pos + RECORD_SIZE
    elif msg_type == 29:
        new_pos = _get_msg29_from_buf(pos, dic)
        warnings.warn("Message 29 encountered, not parsing.", RuntimeWarning)
    elif msg_type == 1:
        new_pos = _get_msg1_from_buf(buf, pos, dic)
    else:   # not message 31 or 1, no decoding performed
        new_pos = pos + RECORD_SIZE

    return new_pos, dic

def _get_msg29_from_buf(pos, dic):
    msg_size = dic['header']['size']
    if msg_size == 65535:
        msg_size = dic['header']['segments'] << 16 | dic['header']['seg_num']
    msg_header_size = _structure_size(MSG_HEADER)
    new_pos = pos + msg_header_size + msg_size
    return new_pos

def _get_msg31_from_buf(buf, pos, dic):
    """ Retrieve and unpack a MSG31 record from a buffer. """
    msg_size = dic['header']['size'] * 2 - 4
    msg_header_size = _structure_size(MSG_HEADER)
    new_pos = pos + msg_header_size + msg_size
    mbuf = buf[pos + msg_header_size:new_pos]
    msg_31_header = _unpack_from_buf(mbuf, 0, MSG_31)
    block_pointers = [v for k, v in msg_31_header.items()
                      if k.startswith('block_pointer') and v > 0]
    for block_pointer in block_pointers:
        block_name, block_dic = _get_msg31_data_block(mbuf, block_pointer)
        dic[block_name] = block_dic

    dic['msg_header'] = msg_31_header
    return new_pos


def _get_msg31_data_block(buf, ptr):
    """ Unpack a msg_31 data block into a dictionary. """
    block_name = buf[ptr + 1: ptr + 4].decode('ascii').strip()

    if block_name == 'VOL':
        dic = _unpack_from_buf(buf, ptr, VOLUME_DATA_BLOCK)
    elif block_name == 'ELV':
        dic = _unpack_from_buf(buf, ptr, ELEVATION_DATA_BLOCK)
    elif block_name == 'RAD':
        dic = _unpack_from_buf(buf, ptr, RADIAL_DATA_BLOCK)
    elif block_name in ['REF', 'VEL', 'SW', 'ZDR', 'PHI', 'RHO', 'CFP']:
        dic = _unpack_from_buf(buf, ptr, GENERIC_DATA_BLOCK)
        ngates = dic['ngates']
        ptr2 = ptr + _structure_size(GENERIC_DATA_BLOCK)
        if dic['word_size'] == 16:
            data = np.frombuffer(buf[ptr2: ptr2 + ngates * 2], '>u2')
        elif dic['word_size'] == 8:
            data = np.frombuffer(buf[ptr2: ptr2 + ngates], '>u1')
        else:
            warnings.warn(
                'Unsupported bit size: %s. Returning array dtype "B"',
                dic['word_size'])
        dic['data'] = data
    else:
        dic = {}
    return block_name, dic


def _get_msg1_from_buf(buf, pos, dic):
    """ Retrieve and unpack a MSG1 record from a buffer. """
    msg_header_size = _structure_size(MSG_HEADER)
    msg1_header = _unpack_from_buf(buf, pos + msg_header_size, MSG_1)
    dic['msg_header'] = msg1_header

    sur_nbins = int(msg1_header['sur_nbins'])
    doppler_nbins = int(msg1_header['doppler_nbins'])

    sur_step = int(msg1_header['sur_range_step'])
    doppler_step = int(msg1_header['doppler_range_step'])

    sur_first = int(msg1_header['sur_range_first'])
    doppler_first = int(msg1_header['doppler_range_first'])
    if doppler_first > 2**15:
        doppler_first = doppler_first - 2**16

    if msg1_header['sur_pointer']:
        offset = pos + msg_header_size + msg1_header['sur_pointer']
        data = np.frombuffer(buf[offset:offset+sur_nbins], '>u1')
        dic['REF'] = {
            'ngates': sur_nbins,
            'gate_spacing': sur_step,
            'first_gate': sur_first,
            'data': data,
            'scale': 2.,
            'offset': 66.,
        }
    if msg1_header['vel_pointer']:
        offset = pos + msg_header_size + msg1_header['vel_pointer']
        data = np.frombuffer(buf[offset:offset+doppler_nbins], '>u1')
        dic['VEL'] = {
            'ngates': doppler_nbins,
            'gate_spacing': doppler_step,
            'first_gate': doppler_first,
            'data': data,
            'scale': 2.,
            'offset': 129.0,
        }
        if msg1_header['doppler_resolution'] == 4:
            # 1 m/s resolution velocity, offset remains 129.
            dic['VEL']['scale'] = 1.
    if msg1_header['width_pointer']:
        offset = pos + msg_header_size + msg1_header['width_pointer']
        data = np.frombuffer(buf[offset:offset+doppler_nbins], '>u1')
        dic['SW'] = {
            'ngates': doppler_nbins,
            'gate_spacing': doppler_step,
            'first_gate': doppler_first,
            'data': data,
            'scale': 2.,
            'offset': 129.0,
        }
    return pos + RECORD_SIZE


def _get_msg5_from_buf(buf, pos, dic):
    """ Retrieve and unpack a MSG1 record from a buffer. """
    msg_header_size = _structure_size(MSG_HEADER)
    msg5_header_size = _structure_size(MSG_5)
    msg5_elev_size = _structure_size(MSG_5_ELEV)

    dic['msg5_header'] = _unpack_from_buf(buf, pos + msg_header_size, MSG_5)
    dic['cut_parameters'] = []
    for i in range(dic['msg5_header']['num_cuts']):
        pos2 = pos + msg_header_size + msg5_header_size + msg5_elev_size * i
        dic['cut_parameters'].append(_unpack_from_buf(buf, pos2, MSG_5_ELEV))
    return pos + RECORD_SIZE


def _structure_size(structure):
    """ Find the size of a structure in bytes. """
    return struct.calcsize('>' + ''.join([i[1] for i in structure]))


def _unpack_from_buf(buf, pos, structure):
    """ Unpack a structure from a buffer. """
    size = _structure_size(structure)
    return _unpack_structure(buf[pos:pos + size], structure)


def _unpack_structure(string, structure):
    """ Unpack a structure from a string. """
    fmt = '>' + ''.join([i[1] for i in structure])  # NEXRAD is big-endian
    lst = struct.unpack(fmt, string)
    return dict(zip([i[0] for i in structure], lst))


# NEXRAD Level II file structures and sizes
# The deails on these structures are documented in:
# "Interface Control Document for the Achive II/User" RPG Build 12.0
# Document Number 2620010E
# and
# "Interface Control Document for the RDA/RPG" Open Build 13.0
# Document Number 2620002M
# Tables and page number refer to those in the second document unless
# otherwise noted.
RECORD_SIZE = 2432
COMPRESSION_RECORD_SIZE = 12
CONTROL_WORD_SIZE = 4

# format of structure elements
# section 3.2.1, page 3-2
CODE1 = 'B'
CODE2 = 'H'
INT1 = 'B'
INT2 = 'H'
INT4 = 'I'
REAL4 = 'f'
REAL8 = 'd'
SINT1 = 'b'
SINT2 = 'h'
SINT4 = 'i'

# Figure 1 in Interface Control Document for the Archive II/User
# page 7-2
VOLUME_HEADER = (
    ('tape', '9s'),
    ('extension', '3s'),
    ('date', 'I'),
    ('time', 'I'),
    ('icao', '4s')
)

# Table II Message Header Data
# page 3-7
MSG_HEADER = (
    ('size', INT2),                 # size of data, no including header
    ('channels', INT1),
    ('type', INT1),
    ('seq_id', INT2),
    ('date', INT2),
    ('ms', INT4),
    ('segments', INT2),
    ('seg_num', INT2),
)

# Table XVII Digital Radar Generic Format Blocks (Message Type 31)
# pages 3-87 to 3-89
MSG_31 = (
    ('id', '4s'),                   # 0-3
    ('collect_ms', INT4),           # 4-7
    ('collect_date', INT2),         # 8-9
    ('azimuth_number', INT2),       # 10-11
    ('azimuth_angle', REAL4),       # 12-15
    ('compress_flag', CODE1),       # 16
    ('spare_0', INT1),              # 17
    ('radial_length', INT2),        # 18-19
    ('azimuth_resolution', CODE1),  # 20
    ('radial_spacing', CODE1),      # 21
    ('elevation_number', INT1),     # 22
    ('cut_sector', INT1),           # 23
    ('elevation_angle', REAL4),     # 24-27
    ('radial_blanking', CODE1),     # 28
    ('azimuth_mode', SINT1),        # 29
    ('block_count', INT2),          # 30-31
    ('block_pointer_1', INT4),      # 32-35  Volume Data Constant XVII-E
    ('block_pointer_2', INT4),      # 36-39  Elevation Data Constant XVII-F
    ('block_pointer_3', INT4),      # 40-43  Radial Data Constant XVII-H
    ('block_pointer_4', INT4),      # 44-47  Moment "REF" XVII-{B/I}
    ('block_pointer_5', INT4),      # 48-51  Moment "VEL"
    ('block_pointer_6', INT4),      # 52-55  Moment "SW"
    ('block_pointer_7', INT4),      # 56-59  Moment "ZDR"
    ('block_pointer_8', INT4),      # 60-63  Moment "PHI"
    ('block_pointer_9', INT4),      # 64-67  Moment "RHO"
    ('block_pointer_10', INT4),     # Moment "CFP"
)


# Table III Digital Radar Data (Message Type 1)
# pages 3-7 to
MSG_1 = (
    ('collect_ms', INT4),           # 0-3
    ('collect_date', INT2),         # 4-5
    ('unambig_range', SINT2),       # 6-7
    ('azimuth_angle', CODE2),       # 8-9
    ('azimuth_number', INT2),       # 10-11
    ('radial_status', CODE2),       # 12-13
    ('elevation_angle', INT2),      # 14-15
    ('elevation_number', INT2),     # 16-17
    ('sur_range_first', CODE2),     # 18-19
    ('doppler_range_first', CODE2),  # 20-21
    ('sur_range_step', CODE2),      # 22-23
    ('doppler_range_step', CODE2),  # 24-25
    ('sur_nbins', INT2),            # 26-27
    ('doppler_nbins', INT2),        # 28-29
    ('cut_sector_num', INT2),       # 30-31
    ('calib_const', REAL4),         # 32-35
    ('sur_pointer', INT2),          # 36-37
    ('vel_pointer', INT2),          # 38-39
    ('width_pointer', INT2),        # 40-41
    ('doppler_resolution', CODE2),  # 42-43
    ('vcp', INT2),                  # 44-45
    ('spare_1', '8s'),              # 46-53
    ('spare_2', '2s'),              # 54-55
    ('spare_3', '2s'),              # 56-57
    ('spare_4', '2s'),              # 58-59
    ('nyquist_vel', SINT2),         # 60-61
    ('atmos_attenuation', SINT2),   # 62-63
    ('threshold', SINT2),           # 64-65
    ('spot_blank_status', INT2),    # 66-67
    ('spare_5', '32s'),             # 68-99
    # 100+  reflectivity, velocity and/or spectral width data, CODE1
)

# Table XI Volume Coverage Pattern Data (Message Type 5 & 7)
# pages 3-51 to 3-54
MSG_5 = (
    ('msg_size', INT2),
    ('pattern_type', CODE2),
    ('pattern_number', INT2),
    ('num_cuts', INT2),
    ('clutter_map_group', INT2),
    ('doppler_vel_res', CODE1),     # 2: 0.5 degrees, 4: 1.0 degrees
    ('pulse_width', CODE1),         # 2: short, 4: long
    ('spare', '10s')                # halfwords 7-11 (10 bytes, 5 halfwords)
)

MSG_5_ELEV = (
    ('elevation_angle', CODE2),  # scaled by 360/65536 for value in degrees.
    ('channel_config', CODE1),
    ('waveform_type', CODE1),
    ('super_resolution', CODE1),
    ('prf_number', INT1),
    ('prf_pulse_count', INT2),
    ('azimuth_rate', CODE2),
    ('ref_thresh', SINT2),
    ('vel_thresh', SINT2),
    ('sw_thresh', SINT2),
    ('zdr_thres', SINT2),
    ('phi_thres', SINT2),
    ('rho_thres', SINT2),
    ('edge_angle_1', CODE2),
    ('dop_prf_num_1', INT2),
    ('dop_prf_pulse_count_1', INT2),
    ('spare_1', '2s'),
    ('edge_angle_2', CODE2),
    ('dop_prf_num_2', INT2),
    ('dop_prf_pulse_count_2', INT2),
    ('spare_2', '2s'),
    ('edge_angle_3', CODE2),
    ('dop_prf_num_3', INT2),
    ('dop_prf_pulse_count_3', INT2),
    ('spare_3', '2s'),
)

# Table XVII-B Data Block (Descriptor of Generic Data Moment Type)
# pages 3-90 and 3-91
GENERIC_DATA_BLOCK = (
    ('block_type', '1s'),
    ('data_name', '3s'),        # VEL, REF, SW, RHO, PHI, ZDR
    ('reserved', INT4),
    ('ngates', INT2),
    ('first_gate', SINT2),
    ('gate_spacing', SINT2),
    ('thresh', SINT2),
    ('snr_thres', SINT2),
    ('flags', CODE1),
    ('word_size', INT1),
    ('scale', REAL4),
    ('offset', REAL4),
    # then data
)

# Table XVII-E Data Block (Volume Data Constant Type)
# page 3-92
VOLUME_DATA_BLOCK = (
    ('block_type', '1s'),
    ('data_name', '3s'),
    ('lrtup', INT2),
    ('version_major', INT1),
    ('version_minor', INT1),
    ('lat', REAL4),
    ('lon', REAL4),
    ('height', SINT2),
    ('feedhorn_height', INT2),
    ('refl_calib', REAL4),
    ('power_h', REAL4),
    ('power_v', REAL4),
    ('diff_refl_calib', REAL4),
    ('init_phase', REAL4),
    ('vcp', INT2),
    ('spare', '2s'),
)

# Table XVII-F Data Block (Elevation Data Constant Type)
# page 3-93
ELEVATION_DATA_BLOCK = (
    ('block_type', '1s'),
    ('data_name', '3s'),
    ('lrtup', INT2),
    ('atmos', SINT2),
    ('refl_calib', REAL4),
)

# Table XVII-H Data Block (Radial Data Constant Type)
# pages 3-93
RADIAL_DATA_BLOCK = (
    ('block_type', '1s'),
    ('data_name', '3s'),
    ('lrtup', INT2),
    ('unambig_range', SINT2),
    ('noise_h', REAL4),
    ('noise_v', REAL4),
    ('nyquist_vel', SINT2),
    ('spare', '2s')
)



"""
Script to check the compliance of a file with the CF/Radial 1.2 standard
"""

import time
import argparse

import numpy as np
import netCDF4

# variable types
DOUBLE = np.float64
FLOAT = np.float32
BYTE = np.byte
INT = np.int32
SHORT = np.int16
STRING = unicode
CHAR = np.dtype('S1')


class AttributeTable:
    """
    A class represeting a table of required/optional variable attributes.
    """

    def __init__(self, text, section):
        self.text = text
        self.section = section
        self.attributes = {}
        self.required_attrs = []
        self.optional_attrs = []

    def add_attr(self, attr_name, required, _type=None, value=None):
        """ Add an attribute to the table. """
        self.attributes[attr_name] = Attribute(_type, value)
        if required:
            self.required_attrs.append(attr_name)
        else:
            self.optional_attrs.append(attr_name)

    def req_attr(self, attr_name, _type=None, value=None):
        """ Add a required attribute to the table. """
        self.attributes[attr_name] = Attribute(_type, value)
        self.required_attrs.append(attr_name)

    def opt_attr(self, attr_name, _type=None, value=None):
        """ Add an optional attribute to the table. """
        self.attributes[attr_name] = Attribute(_type, value)
        self.optional_attrs.append(attr_name)

    def check_attr(self, attr_name, required, test_var, verb):
        """ Check an attribute, log errors or notes if detected. """
        if attr_name not in test_var.ncattrs():
            if required:
                t = "Required %s '%s' missing." % (self.text, attr_name)
                log_error(self.section, t)
                return
            if verb:
                t = "Optional %s '%s' missing." % (self.text, attr_name)
                log_note(self.section, t)
            return

        attr_obj = self.attributes[attr_name]
        attr = getattr(test_var, attr_name)

        # check for incorrect type
        if attr_obj.type_bad(type(attr)):
            tup = (self.text, attr_name, type(attr), attr_obj._type)
            t = "%s '%s' has incorrect type: %s should be %s." % tup
            log_error(self.section, t)

        # check for incorrect value
        if attr_obj.value_bad(attr):
            tup = (self.text, attr_name, attr, attr_obj.value)
            t = "%s '%s' has incorrect value: %s should be %s." % tup
            log_error(self.section, t)

    def check(self, test_var, verb=False):
        """ Check all attributes for errror and notes. """
        # check for required attributes
        for attr_name in self.required_attrs:
            self.check_attr(attr_name, True, test_var, verb)
        for attr_name in self.optional_attrs:
            self.check_attr(attr_name, False, test_var, verb)


class Attribute:
    """
    A class for holding and checking netCDF variable attributes.

    Parameters
    ----------
    _type : type or None
        Expected type of the attribute, None for no expected type.
    value : any or None
        Expected value for the attribute, None for no expected value.

    """

    def __init__(self, _type=None, value=None):
        """ initialize the object. """
        self._type = _type
        self.value = value

    def type_bad(self, _type):
        """
        Return True if the provided type does not matches the expected type.
        """
        if self._type is None or self._type == _type:
            return False
        return True

    def value_bad(self, value):
        """
        Return True if the provided value does not match the expected value.
        """
        if self.value is None or self.value == value:
            return False
        return True


class VariableTable:
    """
    A class represeting a table of required/optional variables.
    """
    def __init__(self, text, section):
        """ Initialize the table. """
        self.text = text
        self.section = section
        self.variables = {}
        self.required_vars = []
        self.optional_vars = []

    def add_var(self, var_name, required, dtype=None, dim=None, units=None):
        """ Add an attribute to the table. """
        self.variables[var_name] = Variable(dtype, dim, units)
        if required:
            self.required_vars.append(var_name)
        else:
            self.optional_vars.append(var_name)

    def req_var(self, var_name, dtype=None, dim=None, units=None):
        """ Add a required variable to the table. """
        self.variables[var_name] = Variable(dtype, dim, units)
        self.required_vars.append(var_name)

    def opt_var(self, var_name, dtype=None, dim=None, units=None):
        """ Add an optional variable to the table. """
        self.variables[var_name] = Variable(dtype, dim, units)
        self.optional_vars.append(var_name)

    def check_var(self, var_name, required, dataset, verb):
        """ Check a specific that a variable exists has correct attributes."""
        if var_name not in dataset.variables:
            if required:
                t = "Required %s '%s' missing." % (self.text, var_name)
                log_error(self.section, t)
                return
            if verb:
                t = "Optional %s '%s' missing." % (self.text, var_name)
                log_note(self.section, t)
            return

        # the variable exist, load the variable and variable object
        var_obj = self.variables[var_name]
        var = dataset.variables[var_name]

        # check for incorrect type
        if var_obj.dtype_bad(var.dtype):
            tup = (self.text, var_name, var.dtype, var_obj.dtype)
            t = "%s '%s' has incorrect type: %s should be %s" % tup
            log_error(self.section, t)
        # check for incorrect dim
        if var_obj.dim_bad(var.dimensions):
            tup = (self.text, var_name, var.dimensions, var_obj.dim)
            t = "%s '%s' has incorrect dimensions: %s should be %s" % tup
            log_error(self.section, t)
        # check for bad units
        if 'units' not in var.ncattrs() and var_obj.units is not None:
            t = "%s '%s' is missing a unit attribute." % (self.text, var_name)
            log_error(self.section, t)
            return
        if var_obj.units is not None and var_obj.units_bad(var.units):
            tup = (self.text, var_name, var.units, var_obj.units)
            t = "%s '%s' has incorrect units: %s should be %s" % tup
            log_error(self.section, t)

    def check(self, dataset, verb=False):
        """ Perform all checks on the variables. """
        # check required variables
        for var_name in self.required_vars:
            self.check_var(var_name, True, dataset, verb)
        # check optional variables
        for var_name in self.optional_vars:
            self.check_var(var_name, False, dataset, verb)
        return


class Variable:
    """
    A class for holding and checking netCDF variables.

    Parameters
    ----------
    dtype : type or None
        Expected dtype of the variable, None for no expected type.
    dim : tuple of str
        Expected dimensions of the variable, None for no expected dimensions.
    units : str
        Expected units attribute of the variable.  None for no expected units.

    """

    def __init__(self, dtype=None, dim=None, units=None):
        """ initalize the object. """
        self.dtype = dtype
        self.dim = dim
        self.units = units

    def dtype_bad(self, dtype):
        """ True if the provided dtype does not match the expected dtype. """
        if self.dtype is None:
            return False
        elif self.dtype == dtype:
            return False
        else:
            return True

    def dim_bad(self, dim):
        """ True if the provided dim does not match the expected dim. """
        if self.dim is None:
            return False
        elif self.dim == dim:
            return False
        else:
            return True

    def units_bad(self, units):
        """ True is the provided units does not match the expected units. """
        if self.units is None:
            return False
        elif self.units == units:
            return False
        else:
            return True


def log_error(section, text):
    """
    Log an error, print it to standard out.
    """
    print "ERROR: (%s) %s" % (section, text)


def log_note(section, text):
    """
    Log a note (unmet optional part of the standard), print to standard out.
    """
    print "NOTE: (%s) %s" % (section, text)


def check_attribute(section, obj, text, attr_name, valid_choices):
    """ Checks an attribute which has a set number of valid values. """
    if attr_name in obj.ncattrs():
        attr = getattr(obj, attr_name)
        if attr not in valid_choices:
            tup = (text, attr_name, attr, ' '.join(valid_choices))
            t = "%s '%s' has an invalid value: %s must be one of %s" % tup
            log_error(section, t)
    return


def check_char_variable(section, dataset, text, var_name, valid_options):
    """ Check a char variable which has a set number of valid values."""
    if var_name in dataset.variables:
        var = dataset.variables[var_name][:]
        value = var.tostring().strip('\x00').strip()
        if value not in valid_options:
            tup = (text, var_name, value, ' '.join(valid_options))
            t = "%s '%s' has an invalid value: %s must be one of %s" % tup
            log_error(section, t)
    return


def check_chararr_variable(section, dataset, text, var_name, valid_options):
    """
    Check a variable array of chars which has a set number of valid values.
    """
    if var_name in dataset.variables:
        for i, chars in enumerate(dataset.variables[var_name]):

            value = chars.tostring().strip('\x00').strip()
            if value not in valid_options:
                tup = (text, var_name, i, value, ' '.join(valid_options))
                t = ("%s '%s' has an invalid value in position %i: "
                     "%s must be one of %s" % tup)
                log_error(section, t)
    return


def check_valid_time_format(section, dataset, text, var_name):
    """ Check that a variable is a valid UTC time formatted string. """
    if var_name in dataset.variables:
        s = str(netCDF4.chartostring(dataset.variables[var_name][:]))
        try:
            time.strptime(s, '%Y-%m-%dT%H:%M:%SZ')
        except:
            tup = (text, var_name, s, 'yyyy-mm-ddThh:mm:ssZ')
            t = "%s '%s' has an invalid format: %s should be %s" % (tup)
            log_error(section, t)


def check_metagroup(section, dataset, meta_group_name, valid_meta_group_vars):
    """ Check the meta_group attributes of a meta group. """
    # check that all variable present have a meta_group attribute and it is
    # set correctly.
    for var_name in valid_meta_group_vars:
        if var_name in dataset.variables:
            var = dataset.variables[var_name]
            if 'meta_group' not in var.ncattrs():
                tup = (meta_group_name, var_name)
                text = "%s %s does not have a `meta_group` attribute" % (tup)
                log_error(section, text)
            else:
                if var.meta_group != meta_group_name:
                    tup = (meta_group_name, var_name, var.meta_group,
                           meta_group_name)
                    text = ("%s %s 'meta_group' attribute has incorrect "
                            "value: %s should be %s" % (tup))
                    log_error(section, text)

    # check if other variables have meta_group attribute set to the
    # current meta_group.
    # XXX this might not be considered an error but rather a note
    for var_name in find_all_meta_group_vars(dataset, meta_group_name):
        if var_name not in valid_meta_group_vars:
            text = ('variable %s should not have its meta_group attribute '
                    "set to '%s'" % (var_name, meta_group_name))
            log_error(section, text)


def find_all_meta_group_vars(dataset, meta_group_name):
    """
    Return a list of all variables which are in a given meta_group.
    """
    return [k for k, v in dataset.variables.items() if
            'meta_group' in v.ncattrs() and v.meta_group == meta_group_name]


def check_cfradial_compliance(dataset, verb=False):
    """
    Check a netcdf dataset for CF/Radial compliance.

    Parameters
    ----------
    dataset : Dataset
        NetCDF Dataset to check against CF/Radial version 1.2 standard
    verb : bool
        True to turn on verbose messages (Notes on missing optional
        variables/attributes).  False (default) to suppress

    """
    ##########################
    # 3 Convention hierarchy #
    ##########################
    if 'Conventions' in dataset.ncattrs():
        if 'CF/Radial' not in dataset.Conventions:
            text = "Convention attribute does not contain 'CF/Radial'"
            log_error('3', text)

    #################################
    # 4: CF/Radial base conventions #
    #################################

    # 4.1 Global attributes
    global_attrs = AttributeTable('global attribute', '4.1')
    global_attrs.req_attr('Conventions', STRING)
    global_attrs.opt_attr('version', STRING)
    global_attrs.req_attr('title', STRING)
    global_attrs.req_attr('institution', STRING)
    global_attrs.req_attr('references', STRING)
    global_attrs.req_attr('source', STRING)
    global_attrs.req_attr('history', STRING)
    global_attrs.req_attr('comment', STRING)
    global_attrs.req_attr('instrument_name', STRING)
    global_attrs.opt_attr('site_name', STRING)
    global_attrs.opt_attr('scan_name', STRING)
    global_attrs.opt_attr('scan_id', INT)
    global_attrs.opt_attr('platform_is_mobile', STRING)
    global_attrs.opt_attr('n_gates_vary', STRING)
    global_attrs.check(dataset, verb)

    check_attribute('4.1', dataset, 'global attribute', 'platform_is_mobile',
                    ['true', 'false'])

    check_attribute('4.1', dataset, 'global attribute', 'n_gates_vary',
                    ['true', 'false'])

    n_gates_vary_flag = False
    if 'n_gates_vary' in dataset.ncattrs()and dataset.n_gates_vary == 'true':
        n_gates_vary_flag = True

    mobile_platform = False
    if 'platform_is_mobile' in dataset.ncattrs():
        if dataset.platform_is_mobile == 'true':
            mobile_platform = True

    # 4.2 Dimensions
    section = '4.2'

    if 'time' not in dataset.dimensions:
        log_error(section, "Required dimension 'time' missing.")
    if 'range' not in dataset.dimensions:
        log_error(section, "Required dimension 'range' missing.")
    if 'sweep' not in dataset.dimensions:
        log_error(section, "Required dimension 'sweep' missing.")
    if n_gates_vary_flag and 'n_points' not in dataset.dimensions:
        text = "Dimension 'n_points' is required and missing."
        log_error(section, text)
    if n_gates_vary_flag is False and 'n_points' in dataset.dimensions:
        text = "Dimension 'n_points must not be included."
        log_error(section, text)
    if verb:
        if 'frequency' not in dataset.dimensions:
            log_note(section, "Optional dimension 'frequency' missing")

    # 4.3 Global variables
    global_vars = VariableTable('global variable', '4.3')
    global_vars.req_var('volume_number', INT)
    global_vars.opt_var('platform_type', CHAR)
    global_vars.opt_var('instrument_type', CHAR)
    global_vars.opt_var('primary_axis', CHAR)
    global_vars.req_var('time_coverage_start', CHAR)
    global_vars.req_var('time_coverage_end', CHAR)
    global_vars.opt_var('time_reference', CHAR)
    global_vars.check(dataset, verb)

    valid_platform_types = [
        'fixed', 'vehicle', 'ship', 'aircraft', 'aircraft_fore',
        'aircraft_aft', 'aircraft_tail', 'aircraft_belly', 'aircraft_roof',
        'aircraft_nose', 'satellite_orbit', 'satellite_geostat']
    check_char_variable('4.3', dataset, 'global variable', 'platform_type',
                        valid_platform_types)

    check_char_variable('4.3', dataset, 'global variable', 'instrument_type',
                        ['radar', 'lidar'])

    check_char_variable('4.3', dataset, 'global variable', 'primary_axis',
                        ['axis_z', 'axis_y', 'axis_x'])

    check_valid_time_format('4.3', dataset, 'global_variable',
                            'time_coverage_start')

    check_valid_time_format('4.3', dataset, 'global_variable',
                            'time_coverage_end')

    check_valid_time_format('4.3', dataset, 'global_variable',
                            'time_reference')

    # 4.4 Coordinate variables
    coordinate_vars = VariableTable('coordinate variable', '4.4')
    coordinate_vars.req_var('time', DOUBLE, ('time', ))
    coordinate_vars.req_var('range', FLOAT, ('range', ), 'meters')
    coordinate_vars.check(dataset, verb)

    # 4.4.1 Attributes for time coordinate variable
    time_attrs = AttributeTable('time attribute', '4.4.1')
    time_attrs.req_attr('standard_name', STRING, 'time')
    time_attrs.req_attr('long_name', STRING,
                        'time_in_seconds_since_volume_start')
    time_attrs.req_attr('units', STRING)
    if 'time' in dataset.variables:
        time_attrs.check(dataset.variables['time'], verb)

    # time unit must be 'seconds since yyyy-mm-ddThh:mm:ssZ'
    if 'time' in dataset.variables:
        if 'units' in dataset.variables['time'].ncattrs():
            time_units = dataset.variables['time'].units

            # begins with 'seconds since '
            if not time_units.startswith('seconds since '):
                t = ("time attribute 'units' has an invalid format: " +
                     str(time_units) + "should begin with 'seconds since '")
                log_error('4.4.1', t)

            # end with yyy-mm-ddThh:mm:ssZ
            time_str = time_units[-20:]
            try:
                time.strptime(time_str, '%Y-%m-%dT%H:%M:%SZ')
            except:
                tup = (time_str, 'yyyy-mm-ddThh:mm:ssZ')
                t = ("'time' attribute 'units' has an invalid formatted time"
                     "value: %s should be %s" % (tup))
                log_error('4.4.1', t)

            # and must match time_reference or time_coverage_start
            if 'time_reference' in dataset.variables:
                v = dataset.variables['time_reference']
                s = str(netCDF4.chartostring(v[:]))
                if s != time_str:
                    tup = (time_str, s)
                    t = ("time attribute 'units' does not match time in "
                         "time_reference variable: %s verses %s" % (tup))
                    log_error('4.4.1', t)
            elif 'time_coverage_start' in dataset.variables:
                v = dataset.variables['time_coverage_start']
                s = str(netCDF4.chartostring(v[:]))
                if s != time_str:
                    tup = (time_str, s)
                    t = ("time attribute 'units' does not match time in "
                         "time_coverage_start variable: %s verses %s" % (tup))
                    log_error('4.4.1', t)

    # 4.4.2 Attribute for range coordinate variables
    range_attrs = AttributeTable('range attribute', '4.4.2')
    range_attrs.req_attr('standard_name', STRING,
                         'projection_range_coordinate')
    range_attrs.req_attr('long_name', STRING, 'range_to_measurement_volume')
    range_attrs.req_attr('units', STRING, 'meters')
    range_attrs.req_attr('spacing_is_constant', STRING)
    range_attrs.req_attr('meters_to_center_of_first_gate', FLOAT)
    range_attrs.opt_attr('meters_between_gates', FLOAT)
    range_attrs.req_attr('axis', STRING, 'radial_range_coordinate')
    if 'range' in dataset.variables:
        range_var = dataset.variables['range']
        range_attrs.check(range_var, verb)
        check_attribute('4.4.2', range_var, 'range attribute',
                        'spacing_is_constant', ['true', 'false'])

    # 4.5 Ray dimension variables
    raydim_vars = VariableTable('ray dimension variable', '4.5')
    raydim_vars.req_var('ray_n_gates', INT, ('time', ))
    raydim_vars.req_var('ray_start_index', INT, ('time', ))

    if n_gates_vary_flag:
        raydim_vars.check(dataset)
    else:
        if 'ray_n_gates' in dataset.variables:
            t = "ray dimension variable 'ray_n_gates' must be exist."
            log_error('4.5', t)
        if 'ray_start_index' in dataset.variables:
            t = "ray dimension variable 'ray_start_index' must be exist."
            log_error('4.5', t)

    # 4.6 Location variables
    dim = ()
    if 'platform_is_mobile' in dataset.ncattrs():
        if dataset.platform_is_mobile == 'true':
            dim = ('time', )
    location_vars = VariableTable('location variable', '4.6')
    location_vars.req_var('latitude', DOUBLE, dim, 'degrees_north')
    location_vars.req_var('longitude', DOUBLE, dim, 'degrees_east')
    location_vars.req_var('altitude', DOUBLE, dim, 'meters')
    location_vars.opt_var('altitude_agl', DOUBLE, dim, 'meters')
    location_vars.check(dataset, verb)

    # 4.7 Sweep variables
    sweep_vars = VariableTable('sweep variable', '4.7')
    sweep_vars.req_var('sweep_number', INT, ('sweep', ))
    sweep_vars.req_var('sweep_mode', CHAR)
    sweep_vars.req_var('fixed_angle', FLOAT, ('sweep', ), 'degrees')
    sweep_vars.req_var('sweep_start_ray_index', INT, ('sweep', ))
    sweep_vars.req_var('sweep_end_ray_index', INT, ('sweep', ))
    sweep_vars.opt_var('target_scan_rate', FLOAT, ('sweep', ),
                       'degrees_per_second')
    sweep_vars.check(dataset, verb)

    valid_sweep_modes = [
        'sector', 'coplane', 'rhi', 'vertical_pointing', 'idle',
        'azimuth_surveillance', 'elevation_surveillance', 'sunscan',
        'pointing', 'manual_ppi', 'manual_rhi']
    check_chararr_variable('4.7', dataset, 'sweep variable', 'sweep_mode',
                           valid_sweep_modes)

    # 4.8 Sensor pointing variables
    sensor_vars = VariableTable('sensor pointing variable', '4.8')
    sensor_vars.req_var('azimuth', FLOAT, ('time', ), 'degrees')
    sensor_vars.req_var('elevation', FLOAT, ('time', ), 'degrees')
    sensor_vars.opt_var('scan_rate', FLOAT, ('time', ), 'degrees_per_second')
    sensor_vars.opt_var('antenna_transition', BYTE, ('time', ))
    sensor_vars.check(dataset, verb)

    # 4.8.1 Attributes for azimuth(time) variable
    azimuth_attrs = AttributeTable('azimuth attribute', '4.8.1')
    azimuth_attrs.req_attr('standard_name', STRING, 'beam_azimuth_angle')
    azimuth_attrs.req_attr('long_name', STRING,
                           'azimuth_angle_from_true_north')
    azimuth_attrs.req_attr('units', STRING, 'degrees')
    azimuth_attrs.req_attr('axis', STRING, 'radial_azimuth_coordinate')
    if 'azimuth' in dataset.variables:
        azimuth_attrs.check(dataset.variables['azimuth'], verb)

    # 4.8.2 Attributes for elevation(time) variable
    elev_attrs = AttributeTable('elevation attribute', '4.8.2')
    elev_attrs.req_attr('standard_name', STRING, 'beam_elevation_angle')
    elev_attrs.req_attr('long_name', STRING,
                        'elevation_angle_from_horizontal_plane')
    elev_attrs.req_attr('units', STRING, 'degrees')
    elev_attrs.req_attr('axis', STRING, 'radial_elevation_coordinate')
    if 'elevation' in dataset.variables:
        elev_attrs.check(dataset.variables['elevation'], verb)

    # 4.9 Moving platform geo-reference variables
    georef_vars = VariableTable('moving platform geo-reference variable',
                                '4.9')
    georef_vars.req_var('heading', FLOAT, ('time', ), 'degrees')
    georef_vars.req_var('roll', FLOAT, ('time', ), 'degrees')
    georef_vars.req_var('pitch', FLOAT, ('time', ), 'degrees')
    georef_vars.req_var('drift', FLOAT, ('time', ), 'degrees')
    georef_vars.req_var('rotation', FLOAT, ('time', ), 'degrees')
    georef_vars.req_var('tilt', FLOAT, ('time', ), 'degrees')
    if mobile_platform:
        georef_vars.check(dataset, verb)
    else:  # fixed platfrom
        for v in ['heading', 'roll', 'pitch', 'drift', 'rotation', 'tilt']:
            if v in dataset.variables:
                t = "variable '%s' must be omitted for fixed platforms" % (v)
                log_error('4.9', t)

    # 4.10 Moments field data variables

    # assume all variables with dimensions ('time', 'range') are field data
    # XXX a better way to do this might be to check against section 6.2
    for var_name, var in dataset.variables.items():
        if var.dimensions == ('time', 'range'):

            # check the data type
            if var.dtype not in [BYTE, SHORT, INT, FLOAT, DOUBLE]:
                tup = (var_name, var.dtype)
                t = "field variable '%s' has invalid type: %s" % (tup)
                log_error('4.10', t)

            # check attributes
            if mobile_platform:
                coordinates_value = ('elevation azimuth range heading roll '
                                     'pitch rotation tilt')
            else:
                coordinates_value = 'elevation azimuth range'

            # TODO check standard_name, against variable name
            # TODO check units correct for given standard_name
            text = "field variable %s" % var_name
            field_attrs = AttributeTable(text, '4.10')
            field_attrs.opt_attr('long_name', STRING)
            field_attrs.req_attr('standard_name', STRING)
            field_attrs.req_attr('units', STRING)
            field_attrs.req_attr('_FillValue')
            if var.dtype in [BYTE, SHORT, INT]:
                field_attrs.req_attr('scale_factor', FLOAT)
                field_attrs.req_attr('add_offset', FLOAT)
            field_attrs.req_attr('coordinates', STRING, coordinates_value)
            field_attrs.check(var, verb)

    #####################
    # 5 Sub-conventions #
    #####################

    # 5.1 The instrument_parameters sub-convention
    ip_vars = VariableTable('instrument_parameters variable', '5.1')
    ip_vars.opt_var('frequency', FLOAT, ('frequency', ), 's-1')
    ip_vars.opt_var('follow_mode', CHAR)
    ip_vars.opt_var('pulse_width', FLOAT, ('time', ), 'seconds')
    ip_vars.opt_var('prt_mode', CHAR)
    ip_vars.opt_var('prt', FLOAT, ('time', ), 'seconds')
    ip_vars.opt_var('prt_ratio', FLOAT, ('time', ))
    ip_vars.opt_var('polarization_mode', CHAR)
    ip_vars.opt_var('nyquist_velocity', FLOAT, ('time', ), 'meters_per_second')
    ip_vars.opt_var('unambiguous_range', FLOAT, ('time', ), 'meters')
    ip_vars.opt_var('n_samples', INT, ('time', ))
    ip_vars.opt_var('sampling_ratio', FLOAT, ('time'))
    ip_vars.check(dataset, verb)

    # first dimension should be sweep for _modes variables
    for v in ['follow_mode', 'prt_mode', 'polarization_mode']:
        if v in dataset.variables:
            dim_0 = dataset.variables[v].dimensions[0]
            if dim_0 != 'sweep':
                text = ("instrument_parameters %s must have a first "
                        "dimension of sweep, not %s" % (v, dim_0))
                log_error('5.1', text)

    # check valid options for _modes variables
    valid_follow_modes = [
        'none', 'sun', 'vehicle', 'aircraft', 'target', 'manual']
    check_chararr_variable('5.1', dataset, 'instrument_parameters',
                           'follow_mode', valid_follow_modes)

    valid_prt_modes = ['fixed', 'staggered', 'dual']
    check_chararr_variable('5.1', dataset, 'instrument_parameters',
                           'prt_mode', valid_prt_modes)

    valid_polarization_modes = [
        'horizontal', 'vertical', 'hv_alt', 'hv_sim', 'circular']
    check_chararr_variable('5.1', dataset, 'instrument_parameters',
                           'polarization_mode', valid_polarization_modes)

    # check that meta_group attribute is correctly set.
    valid_ip_vars = [
        'frequency', 'follow_mode', 'pulse_width', 'prt_mode', 'prt',
        'prt_ratio', 'polarization_mode', 'nyquist_velocity',
        'unambiguous_range', 'n_samples', 'sampling_ratio']
    check_metagroup('5.1', dataset, 'instrument_parameters', valid_ip_vars)

    # 5.2 The radar_parameters sub-convention
    rp_vars = VariableTable('radar_parameters', '5.2')
    rp_vars.opt_var('radar_antenna_gain_h', FLOAT, (), 'dB')
    rp_vars.opt_var('radar_antenna_gain_v', FLOAT, (), 'dB')
    rp_vars.opt_var('radar_beam_width_h', FLOAT, (), 'degrees')
    rp_vars.opt_var('radar_beam_width_v', FLOAT, (), 'degrees')
    rp_vars.opt_var('radar_receiver_bandwidth', FLOAT, (), 's-1')
    rp_vars.opt_var('radar_measured_transmit_power_h', FLOAT, ('time', ),
                    'dBm')
    rp_vars.opt_var('radar_measured_transmit_power_b', FLOAT, ('time', ),
                    'dBm')
    rp_vars.check(dataset, verb)

    valid_rp_vars = [
        'radar_antenna_gain_h', 'radar_antenna_gain_v',
        'radar_beam_width_h', 'radar_beam_width_v',
        'radar_receiver_bandwidth',
        'radar_measured_transmit_power_h', 'radar_measured_transmit_power_b']
    check_metagroup('5.2', dataset, 'radar_parameters', valid_rp_vars)

    # 5.3 The lidar_parameter sub-convention
    lp_vars = VariableTable('lidar_parameters', '5.3')
    lp_vars.opt_var('lidar_beam_divergence', FLOAT, (), 'milliradians')
    lp_vars.opt_var('lidar_field_of_view', FLOAT, (), 'milliradians')
    lp_vars.opt_var('lidar_aperature_diameter', FLOAT, (), 'cm')
    lp_vars.opt_var('lidar_aperture_efficiency', FLOAT, (), 'percent')
    lp_vars.opt_var('lidar_peak_power', FLOAT, (), 'watts')
    lp_vars.opt_var('lidar_pulse_energy', FLOAT, (), 'joules')
    lp_vars.check(dataset, verb)

    valid_lp_vars = [
        'lidar_beam_divergence', 'lidar_field_of_view',
        'lidar_aperature_diameter', 'lidar_aperture_efficiency'
        'lidar_peak_power', 'lidar_pulse_energy']
    check_metagroup('5.3', dataset, 'lidar_parameters', valid_lp_vars)

    # 5.4 The radar_calibration sub-convention
    valid_rc_vars = [
        'r_calib_index',
        'r_calib_time',
        'r_calib_pulse_width',
        'r_calib_ant_gain_h',
        'r_calib_ant_gain_v',
        'r_calib_xmit_power_h',
        'r_calib_xmit_power_v',
        'r_calib_two_way_waveguide_loss_h',
        'r_calib_two_way_waveguide_loss_v',
        'r_calib_two_way_radome_loss_h',
        'r_calib_two_way_radome_loss_v',
        'r_calib_receiver_mismatch_loss',
        'r_calib_radar_constant_h',
        'r_calib_radar_constant_v',
        'r_calib_noise_hc',
        'r_calib_noise_vc',
        'r_calib_noise_hx',
        'r_calib_noise_vx',
        'r_calib_receiver_gain_hc',
        'r_calib_receiver_gain_vc',
        'r_calib_receiver_gain_hx',
        'r_calib_receiver_gain_vx',
        'r_calib_base_dbz_1km_hc',
        'r_calib_base_dbz_1km_vc',
        'r_calib_base_dbz_1km_hx',
        'r_calib_base_dbz_1km_vx',
        'r_calib_sun_power_hc',
        'r_calib_sun_power_vc',
        'r_calib_sun_power_hx',
        'r_calib_sun_power_vx',
        'r_calib_noise_source_power_h',
        'r_calib_noise_source_power_v',
        'r_calib_power_measure_loss_h',
        'r_calib_power_measure_loss_v',
        'r_calib_coupler_forward_loss_h',
        'r_calib_coupler_forward_loss_v',
        'r_calib_zdr_correction',
        'r_calib_ldr_correction_h',
        'r_calib_ldr_correction_v',
        'r_calib_system_phidp',
        'r_calib_test_power_h',
        'r_calib_test_power_v',
        'r_calib_receiver_slope_hc',
        'r_calib_receiver_slope_vc',
        'r_calib_receiver_slope_hx',
        'r_calib_receiver_slope_vx', ]
    check_metagroup('5.4', dataset, 'radar_calibration', valid_rc_vars)

    # 5.4.1 Dimensions

    # 5.4.2 Variables
    d = ('r_calib', )
    rc_vars = VariableTable('radar_calibration', '5.4.2')
    rc_vars.opt_var('r_calib_index', BYTE, ('time', ))
    rc_vars.opt_var('r_calib_time', CHAR)
    rc_vars.opt_var('r_calib_pulse_width', FLOAT, d, 'seconds')
    rc_vars.opt_var('r_calib_ant_gain_h', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_ant_gain_v', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_xmit_power_h', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_xmit_power_v', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_two_way_waveguide_loss_h', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_two_way_waveguide_loss_v', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_two_way_radome_loss_h', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_two_way_radome_loss_v', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_receiver_mismatch_loss', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_radar_constant_h', FLOAT, d, 'dB')  # m/mW
    rc_vars.opt_var('r_calib_radar_constant_v', FLOAT, d, 'dB')  # m/mW
    rc_vars.opt_var('r_calib_noise_hc', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_noise_vc', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_noise_hx', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_noise_vx', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_receiver_gain_hc', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_receiver_gain_vc', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_receiver_gain_hx', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_receiver_gain_vx', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_base_dbz_1km_hc', FLOAT, d, 'dBZ')
    rc_vars.opt_var('r_calib_base_dbz_1km_vc', FLOAT, d, 'dBZ')
    rc_vars.opt_var('r_calib_base_dbz_1km_hx', FLOAT, d, 'dBZ')
    rc_vars.opt_var('r_calib_base_dbz_1km_vx', FLOAT, d, 'dBZ')
    rc_vars.opt_var('r_calib_sun_power_hc', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_sun_power_vc', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_sun_power_hx', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_sun_power_vx', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_noise_source_power_h', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_noise_source_power_v', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_power_measure_loss_h', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_power_measure_loss_v', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_coupler_forward_loss_h', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_coupler_forward_loss_v', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_zdr_correction', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_ldr_correction_h', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_ldr_correction_v', FLOAT, d, 'dB')
    rc_vars.opt_var('r_calib_system_phidp', FLOAT, d, 'degrees')
    rc_vars.opt_var('r_calib_test_power_h', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_test_power_v', FLOAT, d, 'dBm')
    rc_vars.opt_var('r_calib_receiver_slope_hc', FLOAT, d)
    rc_vars.opt_var('r_calib_receiver_slope_vc', FLOAT, d)
    rc_vars.opt_var('r_calib_receiver_slope_hx', FLOAT, d)
    rc_vars.opt_var('r_calib_receiver_slope_vx', FLOAT, d)
    rc_vars.check(dataset, verb)

    if 'r_calib_time' in dataset.variables:
        r_calib_time = dataset.variables['r_calib_time']
        dim_0 = r_calib_time.dimensions[0]
        if dim_0 != 'r_calib':
            text = ('radar_calibration r_calib_time must have first '
                    "dimension of 'r_calib' not '%s'" % (dim_0))
            log_error('5.4.2', text)
        else:
            for i, time_arr in enumerate(r_calib_time):
                s = time_arr.tostring().strip('\x00').strip()
                try:
                    time.strptime(s, '%Y-%m-%dT%H:%M:%SZ')
                except:
                    tup = (i, s, 'yyyy-mm-ddThh:mm:ssZ')
                    t = ("radar_calibration r_calib_time has an invalid time "
                         "format in position %i: %s must be %s" % (tup))
                    log_error('5.4.2', t)

    # 5.5 The lidar_calibration sub-convention
    # Not yet defined in the standard

    # 5.6 The platform_velocity sub-convention
    d = ('time', )
    pv_vars = VariableTable('platform_velocity', '5.6')
    pv_vars.req_var('eastward_velocity', FLOAT, d, 'meters_per_second')
    pv_vars.req_var('northward_velocity', FLOAT, d, 'meters_per_second')
    pv_vars.req_var('vertical_velocity', FLOAT, d, 'meters_per_second')
    pv_vars.req_var('eastward_wind', FLOAT, d, 'meters_per_second')
    pv_vars.req_var('northward_wind', FLOAT, d, 'meters_per_second')
    pv_vars.req_var('vertical_wind', FLOAT, d, 'meters_per_second')
    pv_vars.req_var('heading_rate', FLOAT, d, 'degrees_per_second')
    pv_vars.req_var('roll_rate', FLOAT, d, 'degrees_per_second')
    pv_vars.req_var('pitch_rate', FLOAT, d, 'degrees_per_second')

    valid_pv_vars = [
        'eastward_velocity', 'northward_velocity', 'vertical_velocity',
        'eastward_wind', 'northward_wind', 'vertical_wind',
        'heading_rate', 'roll_rate',  'pitch_rate']

    if mobile_platform:
        pv_vars.check(dataset, verb)
        check_metagroup('5.6', dataset, 'platform_velocity', valid_pv_vars)
    else:
        for var_name in valid_pv_vars:
            if var_name in dataset.variables:
                t = ('variable %s should not exist as the platform is'
                     'stationary' % (var_name))
                log_error('5.6', t)

    # 5.7 The geometry_correction sub-convention
    gc_vars = VariableTable('geometry_correction', '5.7')
    gc_vars.opt_var('azimuth_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('elevation_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('range_correction', FLOAT, (), 'meters')
    gc_vars.opt_var('longitude_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('latitude_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('pressure_altitude_correction', FLOAT, (), 'meters')
    gc_vars.opt_var('radar_altitude_correction', FLOAT, (), 'meters')
    gc_vars.opt_var('eastward_ground_speed_correction', FLOAT, (),
                    'meter_per_second')
    gc_vars.opt_var('northward_ground_speed_correction', FLOAT, (),
                    'meter_per_second')
    gc_vars.opt_var('vertical_velocity_correction', FLOAT, (),
                    'meter_per_second')
    gc_vars.opt_var('heading_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('roll_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('pitch_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('drift_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('rotation_correction', FLOAT, (), 'degrees')
    gc_vars.opt_var('tilt_correction', FLOAT, (), 'degrees')
    gc_vars.check(dataset, verb)

    valid_gc_vars = [
        'azimuth_correction',
        'elevation_correction',
        'range_correction',
        'longitude_correction',
        'latitude_correction',
        'pressure_altitude_correction',
        'radar_altitude_correction',
        'eastward_ground_speed_correction',
        'northward_ground_speed_correction',
        'vertical_velocity_correction',
        'heading_correction',
        'roll_correction',
        'pitch_correction',
        'drift_correction',
        'rotation_correction',
        'tilt_correction']
    check_metagroup('5.7', dataset, 'geometry_correction', valid_gc_vars)

    ####################
    # 6 Standard Names #
    ####################

    # 6.1 Proposed standard names for metadata variables

    # 6.2 Standard names for moments variables
    return


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Check a file for CF/Radial version 1.2 compliance.')
    parser.add_argument('filename', type=str, help='netcdf file to check')
    parser.add_argument('--verb', '-v', dest='verb', action='store_true',
                        default=False, help='turn on verbose messages')
    args = parser.parse_args()

    dataset = netCDF4.Dataset(args.filename, 'r')
    check_cfradial_compliance(dataset, args.verb)



"""
Utilities for reading gamic hdf5 files.

"""

# TODO to move out of aux_io namespace:
# * unit tests - need small sample files.
# * auto-detect file type with pyart.io.read function
# * move to pyart.io namespace

import datetime
import warnings

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar
try:
    from .gamicfile import GAMICFile
    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False
from ..exceptions import MissingOptionalDependency


LIGHT_SPEED = 2.99792458e8  # speed of light in meters per second


def read_gamic(filename, field_names=None, additional_metadata=None,
               file_field_names=False, exclude_fields=None,
               include_fields=None, valid_range_from_file=True,
               units_from_file=True, pulse_width=None, **kwargs):
    """
    Read a GAMIC hdf5 file.

    Parameters
    ----------
    filename : str
        Name of GAMIC HDF5 file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    valid_range_from_file : bool, optional
        True to extract valid range (valid_min and valid_max) for all
        field from the file when they are present.  False will not extract
        these parameters.
    units_from_file : bool, optional
        True to extract the units for all fields from the file when available.
        False will not extract units using the default units for the fields.
    pulse_width : list or None,
        Mandatory for gamic radar processors which have pulsewidth enums.
        pulse_width should contain the pulsewidth' in us.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_gamic but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('gamic', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # Open HDF5 file and get handle
    gfile = GAMICFile(filename)

    # verify that all scans are present in file
    assert gfile.is_file_complete()

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    latitude['data'] = gfile.where_attr('lat', 'float64')
    longitude['data'] = gfile.where_attr('lon', 'float64')
    altitude['data'] = gfile.where_attr('height', 'float64')

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'GAMIC-HDF5'
    what_mapping = {
        'version': 'gamic_version', 'sets_scheduled': 'sets_scheduled',
        'object': 'gamic_object', 'date': 'gamic_date'}
    for gamic_key, metadata_key in what_mapping.items():
        if gfile.is_attr_in_group('what', gamic_key):
            metadata[metadata_key] = gfile.raw_group_attr('what', gamic_key)

    how_keys = ['software', 'template_name', 'site_name', 'host_name',
                'azimuth_beam', 'elevation_beam', 'sdp_name', 'sw_version',
                'sdp_version', 'simulated']
    for key in how_keys:
        if gfile.is_attr_in_group('how', key):
            metadata[key] = gfile.raw_group_attr('how', key)

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_start_ray_index['data'] = gfile.start_ray.astype('int32')
    sweep_end_ray_index['data'] = gfile.end_ray.astype('int32')

    # sweep number
    sweep_number = filemetadata('sweep_number')
    try:
        sweep_number['data'] = gfile.what_attrs('set_idx', 'int32')
    except KeyError:
        sweep_number['data'] = np.arange(gfile.nsweeps, dtype='int32')

    # sweep_type
    scan_type = gfile.raw_scan0_group_attr('what', 'scan_type').lower()
    if hasattr(scan_type, 'decode'):
        scan_type = scan_type.decode('utf-8')
    # check that all scans in the volume are the same type
    if not gfile.is_file_single_scan_type():
        raise NotImplementedError('Mixed scan_type volume.')
    if scan_type not in ['ppi', 'rhi']:
        message = "Unknown scan type: %s, reading as RHI scans." % (scan_type)
        warnings.warn(message)
        scan_type = 'rhi'

    # sweep_mode, fixed_angle
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    if scan_type == 'rhi':
        sweep_mode['data'] = np.array(gfile.nsweeps * ['rhi'])
        fixed_angle['data'] = gfile.how_attrs('azimuth', 'float32')
    elif scan_type == 'ppi':
        sweep_mode['data'] = np.array(gfile.nsweeps * ['azimuth_surveillance'])
        fixed_angle['data'] = gfile.how_attrs('elevation', 'float32')

    # time
    time = filemetadata('time')
    t_data = gfile.ray_header('timestamp', 'int64')
    start_epoch = t_data[0] // 1.e6     # truncate to second resolution
    start_time = datetime.datetime.utcfromtimestamp(start_epoch)
    time['units'] = make_time_unit_str(start_time)
    time['data'] = ((t_data - start_epoch * 1.e6) / 1.e6).astype('float64')

    # range
    _range = filemetadata('range')
    ngates = int(gfile.raw_scan0_group_attr('how', 'bin_count'))
    range_start = float(gfile.raw_scan0_group_attr('how', 'range_start'))
    #n_samples insertion
    range_samples = int(gfile.raw_scan0_group_attr('how', 'range_samples')) 
    range_step = float(gfile.raw_scan0_group_attr('how', 'range_step'))*range_samples
    # range_step may need to be scaled by range_samples
    # XXX This gives distances to start of gates not center, this matches
    # Radx but may be incorrect, add range_step / 2. for center
    _range['data'] = (np.arange(ngates, dtype='float32') * range_step +
                      range_start)
    _range['meters_to_center_of_first_gate'] = range_start
    _range['meters_between_gates'] = range_step

    # elevation
    elevation = filemetadata('elevation')
    start_angle = gfile.ray_header('elevation_start', 'float32')
    stop_angle = gfile.ray_header('elevation_stop', 'float32')
    elevation['data'] = _avg_radial_angles(start_angle, stop_angle)

    # azimuth
    azimuth = filemetadata('azimuth')
    start_angle = gfile.ray_header('azimuth_start', 'float32')
    stop_angle = gfile.ray_header('azimuth_stop', 'float32')
    azimuth['data'] = _avg_radial_angles(start_angle, stop_angle) % 360.

    # fields
    fields = {}
    moment_groups = gfile.moment_groups()
    moment_names = gfile.moment_names(moment_groups)

    for moment_name, group in zip(moment_names, moment_groups):

        field_name = filemetadata.get_field_name(moment_name)
        if field_name is None:
            continue

        field_dic = filemetadata(field_name)
        field_dic['data'] = gfile.moment_data(group, 'float32')
        field_dic['_FillValue'] = get_fillvalue()

        if valid_range_from_file:
            try:
                valid_min = gfile.raw_scan0_group_attr(group, 'dyn_range_min')
                valid_max = gfile.raw_scan0_group_attr(group, 'dyn_range_max')
                field_dic['valid_min'] = valid_min.decode('utf-8')
                field_dic['valid_max'] = valid_max.decode('utf-8')
            except:
                pass

        if units_from_file:
            try:
                units = gfile.raw_scan0_group_attr(group, 'unit')
                field_dic['units'] = units.decode('utf-8')
            except:
                pass
        fields[field_name] = field_dic

    # ray_angle_res
    ray_angle_res = filemetadata('ray_angle_res')
    ray_angle_res['data'] = gfile.how_attrs('angle_step', 'float32')

    # rays_are_indexed
    rays_are_indexed = filemetadata('rays_are_indexed')
    rays_are_indexed['data'] = np.array(
        [['false', 'true'][i] for i in gfile.how_attrs('angle_sync', 'uint8')])

    # target_scan_rate
    target_scan_rate = filemetadata('target_scan_rate')
    target_scan_rate['data'] = gfile.how_attrs('scan_speed', 'float32')

    # scan_rate
    scan_rate = filemetadata('scan_rate')
    scan_rate['data'] = target_scan_rate['data']
    if scan_type == 'ppi':
        azs_names = ['az_speed', 'azimuth_speed']
        azs_name = azs_names[0]
        for azs_name in azs_names:
            if gfile.is_field_in_ray_header(azs_name):
                scan_rate['data'] = gfile.ray_header(azs_name, 'float32')
                break
    elif scan_type == 'rhi':
        els_names = ['el_speed', 'elevation_speed']
        els_name = els_names[0]
        for els_name in els_names:
            if gfile.is_field_in_ray_header(els_name):
                scan_rate['data'] = gfile.ray_header(els_name, 'float32')
                break
    else:
        scan_rate = None

    # instrument_parameters
    instrument_parameters = _get_instrument_params(gfile, filemetadata,
                                                   pulse_width)

    gfile.close()

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters,
        ray_angle_res=ray_angle_res,
        rays_are_indexed=rays_are_indexed,
        scan_rate=scan_rate,
        target_scan_rate=target_scan_rate)


def _get_instrument_params(gfile, filemetadata, pulse_width):
    """ Return a dictionary containing instrument parameters. """

    instrument_params = {}

    dic = filemetadata('frequency')
    dic['data'] = np.array(
        [LIGHT_SPEED / gfile.raw_scan0_group_attr('how', 'radar_wave_length')],
        dtype='float32')
    instrument_params['frequency'] = dic

    dic = filemetadata('radar_beam_width_h')
    dic['data'] = gfile.how_attr('azimuth_beam', 'float32')
    instrument_params['radar_beam_width_h'] = dic

    dic = filemetadata('radar_beam_width_v')
    dic['data'] = gfile.how_attr('elevation_beam', 'float32')
    instrument_params['radar_beam_width_v'] = dic

    dic = filemetadata('pulse_width')
    pw_names = ['pulse_width_us', 'pulse_width_mks', 'pulse_width']
    pw_name = 'pulse_width_us'
    dic['data'] = None
    for pw_name in pw_names:
        if gfile.is_attr_in_group('/scan0/how', pw_name):
            if pw_name == 'pulse_width':
                dic['data'] = gfile.sweep_expand(
                    pulse_width[gfile.how_attrs(pw_name, 'int')[0]] * 1e-6)
            else:
                dic['data'] = gfile.sweep_expand(
                    gfile.how_attrs(pw_name, 'float32') * 1e-6)
            break
    instrument_params['pulse_width'] = dic

    dic = filemetadata('prt')
    dic['data'] = gfile.sweep_expand(1. / gfile.how_attrs('PRF', 'float32'))
    instrument_params['prt'] = dic

    unfolding = gfile.how_attrs('unfolding', 'int32')
    dic = filemetadata('prt_mode')
    dic['data'] = np.array([_prt_mode_from_unfolding(i) for i in unfolding])
    instrument_params['prt_mode'] = dic

    dic = filemetadata('prt_ratio')
    dic['data'] = gfile.sweep_expand(
        [[1, 2./3., 3./4., 4./5.][i] for i in unfolding])
    instrument_params['prt_ratio'] = dic

    dic = filemetadata('unambiguous_range')
    dic['data'] = gfile.sweep_expand(gfile.how_attrs('range', 'float32'))
    instrument_params['unambiguous_range'] = dic

    if gfile.is_attr_in_group('/scan0/how/extended', 'nyquist_velocity'):
        dic = filemetadata('nyquist_velocity')
        dic['data'] = gfile.sweep_expand(
            gfile.how_ext_attrs('nyquist_velocity'))
        instrument_params['nyquist_velocity'] = dic

    dic = filemetadata('n_samples')
    dic['data'] = gfile.sweep_expand(
        gfile.how_attrs('range_samples', 'int32') *
        gfile.how_attrs('time_samples', 'int32'), dtype='int32')
    instrument_params['n_samples'] = dic

    return instrument_params


def _avg_radial_angles(angle1, angle2):
    """ Return the average angle between two radial angles. """
    return np.angle(
        (np.exp(1.j*np.deg2rad(angle1)) +
         np.exp(1.j*np.deg2rad(angle2))) / 2., deg=True)


def _prt_mode_from_unfolding(unfolding):
    """ Return 'fixed' or 'staggered' depending on unfolding flag """
    if unfolding == 0:
        return 'fixed'
    else:
        return 'staggered'



"""
Routines for reading GCPEX D3R files.

"""

import datetime

import numpy as np
import netCDF4

from ..config import FileMetadata
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar


D3R_FIELD_NAMES = {
    # corrected reflectivity, horizontal
    'Reflectivity': 'reflectivity',
    # corrected reflectivity, vertical
    'DBZV': 'reflectivity',
    # differential reflectivity
    'DifferentialReflectivity': 'differential_reflectivity',
    'CrossPolCorrelation': 'cross_correlation_ratio',
    'ClutterPowerH': 'clutter_power_h',
    'ClutterPowerV': 'clutter_power_v',
    'DifferentialPhase': 'differential_phase',
    'KDP': 'specific_differential_phase',
    'NormalizedCoherentPower': 'normalized_coherent_power',
    'Signal+Clutter_toNoise_H': 'signal_to_noise_ratio',
    'Velocity': 'velocity',
    'SpectralWidth': 'spectrum_width',
    'SignalPower_H': 'signal_power_h',
}


def read_d3r_gcpex_nc(filename, field_names=None, additional_metadata=None,
                      file_field_names=False, exclude_fields=None,
                      include_fields=None, read_altitude_from_nc=False, **kwargs):
    """
    Read a D3R GCPEX netCDF file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    read_altitude_from_nc : bool, optional
        True if you want the altitude value to be read from the provider netCDF file.
        False will default to the value np.array([295.], dtype='float64')

    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = D3R_FIELD_NAMES
    filemetadata = FileMetadata('cfradial', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # One sweep per file
    nsweeps = 1

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    latitude['data'] = np.array([ncobj.Latitude], dtype='float64')
    longitude['data'] = np.array([ncobj.Longitude], dtype='float64')
    altitude_list = [ncobj.Altitude] if read_altitude_from_nc else [295.]
    altitude['data'] = np.array(altitude_list, dtype='float64')

    # metadata
    metadata = filemetadata('metadata')
    metadata['source'] = "Colorado State EE - chandrasekar"
    metadata['original_container'] = 'D3R_gcpex_nc'
    metadata['nc_conventions'] = ncobj.NetCDFRevision

    metadata['version'] = ncobj.NetCDFRevision
    metadata['source'] = "Chandra"

    metadata['system'] = ncobj.RadarName
    metadata['software'] = ncobj.NetCDFRevision
    metadata['sw_version'] = ncobj.NetCDFRevision

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    rays_per_sweep = np.shape(ncvars['Azimuth'][:])
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
    seri = np.cumsum(rays_per_sweep).astype('int32') - 1
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = seri

    # sweep_number
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')

    # sweep_mode
    sweep_mode = filemetadata('sweep_mode')
    sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])

    # scan_type
    if ncobj.ScanType == 2:
        scan_type = 'ppi'
    else:
        scan_type = 'rhi'

    # fixed_angle
    fixed_angle = filemetadata('fixed_angle')
    if ncobj.ScanType == 2:
        sweep_el = ncvars['Elevation'][0]
    else:
        sweep_el = ncvars['Azimuth'][0]
    fixed_angle['data'] = np.array([sweep_el], dtype='float32')

    # elevation
    elevation = filemetadata('elevation')
    elevation['data'] = ncvars['Elevation']

    # range
    _range = filemetadata('range')
    # check that the gate spacing is constant between sweeps
    rstart = ncvars['StartRange'][:]
    if any(rstart != rstart[0]):
        raise ValueError('range start changes between sweeps')
    rscale = ncvars['GateWidth'][:]/1000.
    if any(rscale != rscale[0]):
        raise ValueError('range scale changes between sweeps')

    nbins = ncobj.NumGates
    _range['data'] = (np.arange(nbins, dtype='float32') * rscale[0] +
                      rstart[0] * 1000.)
    _range['meters_to_center_of_first_gate'] = rstart[0]
    _range['meters_between_gates'] = float(rscale[0])

    # azimuth
    azimuth = filemetadata('azimuth')
    azimuth['data'] = ncvars['Azimuth'][:]

    # time
    _time = filemetadata('time')
    start_time = datetime.datetime.utcfromtimestamp(ncobj.Time)
    _time['units'] = make_time_unit_str(start_time)
    _time['data'] = (ncvars['Time']-ncobj.Time).astype('float32')

    # fields
    # all variables with dimensions of 'Radial', 'Gate' are fields
    keys = [k for k, v in ncvars.items()
            if v.dimensions == ('Radial', 'Gate')]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            if include_fields is not None:
                if not key in include_fields:
                    continue
            field_name = key
        fields[field_name] = _ncvar_to_dict(ncvars[key])

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def _ncvar_to_dict(ncvar):
    """ Convert a NetCDF Dataset variable to a dictionary. """
    # copy all attribute except for scaling parameters
    d = dict((k, getattr(ncvar, k)) for k in ncvar.ncattrs()
             if k not in ['scale_factor', 'add_offset'])
    d['data'] = ncvar[:]
    if np.isscalar(d['data']):
        # netCDF4 1.1.0+ returns a scalar for 0-dim array, we always want
        # 1-dim+ arrays with a valid shape.
        d['data'] = np.array(d['data'])
        d['data'].shape = (1, )
    return d



""" Unit Tests for Py-ART's correct/dealias.py module. """

# python test_dealias.py
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import pytest

import pyart


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_dealias_sounding():
    radar, dealias_vel = perform_dealias()
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    assert dealias_vel['data'][13, 46] is np.ma.masked
    assert dealias_vel['data'][13, 47] is np.ma.masked


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_set_limits():

    radar, dealias_vel = perform_dealias(set_limits=True)
    assert 'valid_min' in dealias_vel
    assert_almost_equal(dealias_vel['valid_min'], -30.0)
    assert 'valid_max' in dealias_vel
    assert_almost_equal(dealias_vel['valid_max'], 30.0)

    radar, dealias_vel = perform_dealias(set_limits=False)
    assert 'valid_min' not in dealias_vel
    assert 'valid_max' not in dealias_vel


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_dealias_sounding_keep_original():
    radar, dealias_vel = perform_dealias(True)
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    assert dealias_vel['data'][13, 46] == -7.5
    assert dealias_vel['data'][13, 47] == 8.5


def perform_dealias(keep_original=False, **kwargs):
    """ Perform velocity dealiasing on reference data. """
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be dealiased.
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    height = np.linspace(150, 250, 10).astype('float32')
    speed = np.ones((10), dtype='float32') * 0.5
    direction = np.ones((10), dtype='float32') * 5.
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, sonde_profile=profile, keep_original=keep_original,
        **kwargs)
    return radar, dealias_vel


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_dealias_last_radar_and_sounding_with_profile():
    radar = pyart.testing.make_velocity_aliased_radar()
    last_radar = pyart.testing.make_velocity_aliased_radar(False)
    height = np.linspace(150, 250, 10).astype('float32')
    speed = np.ones((10), dtype='float32') * 0.5
    direction = np.ones((10), dtype='float32') * 5.
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, sonde_profile=profile, last_radar=last_radar,
        last_vel_field='velocity')
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    return


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_dealias_last_radar_and_sounding():
    radar = pyart.testing.make_velocity_aliased_radar()
    last_radar = pyart.testing.make_velocity_aliased_radar(False)
    dealias_vel = pyart.correct.dealias_fourdd(
        radar, last_radar=last_radar, last_vel_field='velocity')
    assert_allclose(
        dealias_vel['data'][13, :27],
        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
         12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
         2.5, 1.5, 0.5])
    return


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_error_raising():
    # ValueError when no sounding or last_radar provided

    radar = pyart.testing.make_velocity_aliased_radar()
    pytest.raises(ValueError, pyart.correct.dealias_fourdd, radar)
    return


@pytest.mark.skipif(not pyart.correct.dealias._FOURDD_AVAILABLE,
                    reason="TRMM RSL is not installed.")
def test_segmentation_fault_error():
    # See GitHub issue #571

    # ValueError when sounding is > 999
    radar = pyart.testing.make_velocity_aliased_radar()
    height = np.arange(2000)
    speed = np.zeros((2000, ))
    direction = np.zeros((2000, ))
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    pytest.raises(
        ValueError, pyart.correct.dealias_fourdd, radar,
        sonde_profile=profile)
    return


if __name__ == "__main__":

    radar, dealias_vel = perform_dealias()
    radar.fields['dealiased_velocity'] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields['velocity']['data'][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields['dealiased_velocity']['data'][13])

    # create plot
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi('velocity', 0, ax=ax1, colorbar_flag=False,
                title='', vmin=-10, vmax=10)
    rd.plot_ppi('dealiased_velocity', 0, ax=ax2, colorbar_flag=False,
                title='', vmin=-10, vmax=20)
    fig.savefig('dealias_plot.png')



"""
Utilities for reading and writing of MDV grid files.

"""


import datetime
import warnings

from netCDF4 import num2date, date2num
import numpy as np

from ..config import FileMetadata, get_fillvalue, get_metadata
from ..core.grid import Grid
from .common import make_time_unit_str, _test_arguments, prepare_for_read
from ..lazydict import LazyLoadDict
from . import mdv_common


def write_grid_mdv(filename, grid, mdv_field_names=None,
                   field_write_order=None):
    """
    Write grid object to MDV file.

    Create a MDV file containing data from the provided grid instance.

    The MDV file will contain parameters from the 'source' key if contained
    in grid.metadata. If this key or parameters related to the radar location
    and name are not present in the grid a default or sentinel value.
    will be written in the MDV file in the place of the parameter.

    Grid fields will be saved in float32 unless the `_Write_as_dtype` key is
    present.

    Parameters
    ----------
    filename : str or file-like object.
        Filename of MDV file to create. If a file-like object is specified
        data will be written using the write method.
    grid : Grid
        Grid object from which to create MDV file.
    mdv_field_names : dict or None, optional
        Mapping between grid fields and MDV data type names. Field names
        mapped to None or with no mapping will be excluded from
        writing. If None, the same field names will be used.
    field_write_order : list or None, optional
        Order in which grid fields should be written out in the MDV file.
        None, the default, will determine a valid order automatically.

    Notes
    -----
    Do to limitations of the MDV format, not all grid objects are writable.
    To write a grid the following conditions must be satisfied:

        * XY grid must be regular (equal spacing), Z can be irregular.
        * The number of Z levels must not exceed 122.
        * Fields can be encoded in the file using the '_Write_as_dtype' key
          specifying one of 'uint8', 'uint16' or 'float32'. Use the
          'scale_factor' and 'add_offset' keys to specify scaling. Field
          data in the Grid object should be uncompressed, that is to say
          it has had the scaling applied.

    """
    # first of all firm field list
    if field_write_order is None:
        field_write_order = list(grid.fields.keys())

    # remove fields not in mdv_field_names
    if mdv_field_names is not None:
        for ifield, field in enumerate(field_write_order):
            if field not in mdv_field_names or mdv_field_names[field] is None:
                field_write_order.pop(ifield)

    # mount empty mdv file
    grid_shape = grid.fields[field_write_order[0]]['data'].shape
    nz, ny, nx = grid_shape
    nfields = len(field_write_order)
    if nz > mdv_common.MDV_MAX_VLEVELS:
        warnings.warn(('%i vlevels exceed MDV_MAX_VLEVELS = %i. Extra ' +
                       'levels will be ignored') %
                      (nz, mdv_common.MDV_MAX_VLEVELS))
        nz = mdv_common.MDV_MAX_VLEVELS
    mdv = mdv_common.MdvFile(None)
    mdv.field_headers = mdv._get_field_headers(nfields)
    mdv.vlevel_headers = mdv._get_vlevel_headers(nfields)
    mdv.fields_data = [None] * nfields
    mdv.projection = 'flat'

    # fill headers
    d = mdv.master_header
    grid_datetime = _time_dic_to_datetime(grid.time)
    mdv.times['time_centroid'] = grid_datetime
    mdv.times["time_begin"] = grid_datetime
    mdv.times["time_end"] = grid_datetime
    mdv._time_dict_into_header()

    d["data_dimension"] = 3  # XXX are grid's always 3d?
    # =DATA_SYNTHESIS, I don't realy know, so miscellaneous!
    d["data_collection_type"] = 3
    if (grid.z['units'] == 'm' or grid.z['units'] == 'meters'):
        d["native_vlevel_type"] = 4
    elif (grid.z['units'] == '\xc2' or grid.z['units'] == 'degree'):
        d["native_vlevel_type"] = 9
    d["vlevel_type"] = d["native_vlevel_type"]
    d["nfields"] = nfields
    d["max_nx"] = nx
    d["max_ny"] = ny
    d["max_nz"] = nz
    td = datetime.datetime.utcnow() - datetime.datetime(1970, 1, 1, 0, 0)
    d["time_written"] = int(
        round(
            td.microseconds + (
                td.seconds + td.days * 24 * 3600) * 10**6) / 10**6)

    # sensor location using radar_ attribute or origin_ if not available
    if grid.radar_longitude is not None and grid.nradar != 0:
        d["sensor_lon"] = grid.radar_longitude['data'][0]
    else:
        d["sensor_lon"] = grid.origin_longitude['data'][0]

    if grid.radar_latitude is not None and grid.nradar != 0:
        d["sensor_lat"] = grid.radar_latitude['data'][0]
    else:
        d["sensor_lat"] = grid.origin_latitude['data'][0]

    if grid.radar_altitude is not None and grid.nradar != 0:
        d["sensor_alt"] = grid.radar_altitude['data'][0] / 1000.
    else:
        d["sensor_alt"] = grid.origin_altitude['data'][0] / 1000.

    for meta_key, mdv_key in mdv_common.MDV_METADATA_MAP.items():
        if meta_key in grid.metadata:
            d[mdv_key] = grid.metadata[meta_key].encode("ASCII")

    for ifield, field in enumerate(field_write_order):
        d = mdv.field_headers[ifield]
        l = mdv.vlevel_headers[ifield]

        # fill fields_header
        d["nx"] = nx
        d["ny"] = ny
        d["nz"] = nz
        d["proj_type"] = mdv_common.PROJ_FLAT
        try:
            dtype = np.dtype(grid.fields[field]['_Write_as_dtype'])
        except KeyError:    # default to float32 encoding
            dtype = np.float32
        if dtype == np.uint8:
            d["encoding_type"] = mdv_common.ENCODING_INT8
            d["data_element_nbytes"] = 1
        elif dtype == np.uint16:
            d["encoding_type"] = mdv_common.ENCODING_INT16
            d["data_element_nbytes"] = 2
        elif dtype == np.float32 or dtype == np.float64:
            d["encoding_type"] = mdv_common.ENCODING_FLOAT32
            d["data_element_nbytes"] = 4
        else:
            raise TypeError("Unsuported encoding %s, encoding must be "
                            "uint8, uint16 or float32 as specfied by "
                            "the '_Write_as_dtype key" % dtype)
        d["compression_type"] = 3   # zlib

        d["scaling_type"] = 4  # SCALING_SPECIFIED (by the user)
        d["native_vlevel_type"] = mdv.master_header["vlevel_type"]
        d["vlevel_type"] = mdv.master_header["vlevel_type"]
        d["data_dimension"] = 3
        d["proj_origin_lat"] = grid.origin_latitude['data'][0]
        d["proj_origin_lon"] = grid.origin_longitude['data'][0]
        d["grid_dx"] = (grid.x['data'][1] - grid.x['data'][0]) / 1000.
        d["grid_dy"] = (grid.y['data'][1] - grid.y['data'][0]) / 1000.
        d["grid_minx"] = (grid.x['data'][0]) / 1000.
        d["grid_miny"] = (grid.y['data'][0]) / 1000.
        if "scale_factor" in grid.fields[field].keys():
            d["scale"] = grid.fields[field]["scale_factor"]
        if "add_offset" in grid.fields[field].keys():
            d["bias"] = grid.fields[field]["add_offset"]

        # bad_data prioritise _FillValue
        if "_FillValue" in grid.fields[field].keys():
            d["bad_data_value"] = grid.fields[field]["_FillValue"]
        elif "missing_value" in grid.fields[field].keys():
            d["bad_data_value"] = grid.fields[field]["missing_value"]
        else:
            d["bad_data_value"] = get_fillvalue()
        # missing_data prioritise missing_value
        if "missing_value" in grid.fields[field].keys():
            d["missing_data_value"] = grid.fields[field]["missing_value"]
        elif "_FillValue" in grid.fields[field].keys():
            d["missing_data_value"] = grid.fields[field]["_FillValue"]
        else:
            d["missing_data_value"] = get_fillvalue()

        d["min_value"] = np.amax(grid.fields[field]['data'])
        d["max_value"] = np.amin(grid.fields[field]['data'])
        if "standard_name" in grid.fields[field].keys():
            d["field_name_long"] = (
                grid.fields[field]["standard_name"].encode("ASCII"))
        elif "long_name" in grid.fields[field].keys():
            d["field_name_long"] = (
                grid.fields[field]["long_name"].encode("ASCII"))
        if mdv_field_names is not None:
            d["field_name"] = mdv_field_names[field].encode("ASCII")
        else:
            d["field_name"] = field.encode("ASCII")
        if "units" in grid.fields[field].keys():
            d["units"] = grid.fields[field]["units"].encode("ASCII")
        d["transform"] = "none".encode("ASCII")  # XXX not implemented

        # fill vlevels_header
        typ = [0] * 122
        level = [0] * 122
        for iz in range(nz):
            typ[iz] = d["vlevel_type"]
            level[iz] = grid.z["data"][iz] / 1000.
        l["type"] = typ
        l["level"] = level

        # put data to field
        mdv.fields_data[ifield] = grid.fields[field]["data"]

    # write the file
    mdv.write(filename)


def read_grid_mdv(filename, field_names=None, additional_metadata=None,
                  file_field_names=False, exclude_fields=None,
                  delay_field_loading=False, **kwargs):
    """
    Read a MDV file to a Grid Object.

    Parameters
    ----------
    filename : str
        Name of MDV file to read or file-like object pointing to the
        beginning of such a file.
    field_names : dict, optional
        Dictionary mapping MDV data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the grid object. This is applied
        after the `file_field_names` and `field_names` parameters.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed. In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects.

    Returns
    -------
    grid : Grid
        Grid object containing data from MDV file.

    Notes
    -----
    This function can only read cartesian MDV files with fields
    compressed with gzip or zlib. For polar files see
    :py:func:`pyart.io.read_mdv`

    MDV files and Grid object are not fully interchangeable. Specific
    limitation include:

        * All fields must have the same shape and dimensions.
        * All fields must have the same projection.
        * Vlevels types must not vary.
        * Projection must not be PROJ_POLAR_RADAR (9) or PROJ_RHI_RADAR (13).
        * Correct unit in the Z axis are just availible for 'vlevel_type'
          equal to VERT_TYPE_Z(4), VERT_TYPE_ELEV(9), VERT_TYPE_AZ(17),
          VERT_TYPE_PRESSURE(3) and VERT_TYPE_THETA(7).
        * The behavior in cases of 2D data is unknown but most likely will not
          fail.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # XXX add tests for conversion limitations
    # create metadata retrieval object
    filemetadata = FileMetadata('mdv', field_names, additional_metadata,
                                file_field_names, exclude_fields)
    mdv = mdv_common.MdvFile(prepare_for_read(filename))

    # time dictionaries
    units = make_time_unit_str(mdv.times['time_begin'])
    time = get_metadata('grid_time')
    time['data'] = np.array([date2num(mdv.times['time_centroid'], units)])
    time['units'] = units

    # origin dictionaries
    origin_altitude = get_metadata('origin_altitude')
    origin_altitude['data'] = np.array(
        [mdv.master_header["sensor_alt"] * 1000.], dtype='float64')

    origin_latitude = get_metadata('origin_latitude')
    origin_latitude['data'] = np.array(
        [mdv.master_header["sensor_lat"]], dtype='float64')

    origin_longitude = get_metadata('origin_longitude')
    origin_longitude['data'] = np.array(
        [mdv.master_header["sensor_lon"]], dtype='float64')

    # grid coordinate dictionaries
    nz = mdv.master_header["max_nz"]
    ny = mdv.master_header["max_ny"]
    nx = mdv.master_header["max_nx"]
    z_line = mdv.vlevel_headers[0]["level"][0:nz]
    y_start = mdv.field_headers[0]["grid_miny"] * 1000.
    x_start = mdv.field_headers[0]["grid_minx"] * 1000.
    y_step = mdv.field_headers[0]["grid_dy"] * 1000.
    x_step = mdv.field_headers[0]["grid_dx"] * 1000.

    if mdv.field_headers[0]["proj_type"] == mdv_common.PROJ_LATLON:
        # x and y dimensions have units of degrees
        xunits = 'degree_E'
        yunits = 'degree_N'
        y_start = mdv.field_headers[0]["grid_miny"]
        x_start = mdv.field_headers[0]["grid_minx"]
        y_step = mdv.field_headers[0]["grid_dy"]
        x_step = mdv.field_headers[0]["grid_dx"]
    elif (mdv.field_headers[0]["proj_type"] != mdv_common.PROJ_POLAR_RADAR and
          mdv.field_headers[0]["proj_type"] != mdv_common.PROJ_RHI_RADAR):
        xunits = 'm'
        yunits = 'm'

    if mdv.field_headers[0]["vlevel_type"] == 4:  # VERT_TYPE_Z
        zunits = 'm'
        z_line = [e * 1000. for e in z_line]
    elif mdv.field_headers[0]["vlevel_type"] == 9:  # VERT_TYPE_ELEV
        zunits = 'degree'  # elevation
    elif mdv.field_headers[0]["vlevel_type"] == 17:  # VERT_TYPE_AZ
        zunits = 'degree'  # azimuth
    elif mdv.field_headers[0]["vlevel_type"] == 3:  # VERT_TYPE_PRESSURE
        zunits = 'mb'
    elif mdv.field_headers[0]["vlevel_type"] == 7:  # VERT_TYPE_THETA
        zunits = 'kelvin'
    else:
        warnings.warn(("While reading MDV found unexpected 'vlevel_type'" +
                       " (%i), units in the z axis set to 'unknown'") %
                      mdv.field_headers[0]["vlevel_type"])
        zunits = 'unknown'

    x = get_metadata('x')
    x['data'] = np.linspace(x_start, x_start + x_step * (nx-1), nx)
    x['units'] = xunits

    y = get_metadata('y')
    y['data'] = np.linspace(y_start, y_start + y_step * (ny-1), ny)
    y['units'] = yunits

    z = get_metadata('z')
    z['data'] = np.array(z_line, dtype='float64')
    z['units'] = zunits

    # metadata
    metadata = filemetadata('metadata')
    for meta_key, mdv_key in mdv_common.MDV_METADATA_MAP.items():
        metadata[meta_key] = mdv.master_header[mdv_key]

    # fields
    fields = {}
    mdv_fields = mdv._make_fields_list()
    for mdv_field in set(mdv_fields):
        field_name = filemetadata.get_field_name(mdv_field)
        if field_name is None:
            continue
        field_dic = filemetadata(field_name)

        field_dic['_FillValue'] = get_fillvalue()
        dataextractor = mdv_common._MdvVolumeDataExtractor(
            mdv, mdv.fields.index(mdv_field), get_fillvalue(), two_dims=False)
        if delay_field_loading:
            field_dic = LazyLoadDict(field_dic)
            field_dic.set_lazy('data', dataextractor)
        else:
            field_dic['data'] = dataextractor()
        fields[field_name] = field_dic

    if not delay_field_loading:
        mdv.close()
    return Grid(time, fields, metadata,
                origin_latitude, origin_longitude, origin_altitude, x, y, z)


# This function may be helpful in other cases and could be moved in common
def _time_dic_to_datetime(dic):
    """ Return a datetime for the first element in a time dictionary. """
    if 'calendar' in dic:
        calendar = dic['calendar']
    else:
        calendar = 'standard'
    return num2date(dic['data'][0], dic['units'], calendar)



import pkg_resources
import pooch

DATASETS = pooch.create(
    path=pooch.os_cache('pyart-datasets'),
    base_url='https://adc.arm.gov/pyart/example_data/',
    env='PYART_DATASETS_DIR',
)


with pkg_resources.resource_stream('pyart.testing', 'registry.txt') as registry_file:
    DATASETS.load_registry(registry_file)


def locate():
    """The absolute path to the sample data storage location on disk.
    This is where the data are saved on your computer. The location is
    dependent on the operating system. The folder locations are defined by the
    ``appdirs``  package (see the `appdirs documentation
    <https://github.com/ActiveState/appdirs>`__).
    The location can be overwritten by the ``PYTHIA_DATASETS_DIR`` environment
    variable to the desired destination.
    Returns
    -------
    path : str
        The local data storage location.
    """
    return str(DATASETS.abspath)


def get_test_data(file):
    """
    Accesses the desired test data storage
    Parameters
    ----------
    file = str
        The name of the desired file
    Returns
    -------
    path_to_file = str
        Local path to the desired file
    """
    return DATASETS.fetch(file)


import pyart
import numpy as np

from copy import deepcopy


def test_gatemapper():
    # Make fake radar with target
    old_radar = pyart.testing.make_target_radar()
    new_radar = deepcopy(old_radar)
    new_radar.latitude["data"] = old_radar.latitude["data"] + 0.001
    new_radar.longitude["data"] = old_radar.longitude["data"] + 0.001
    gate_mapper = pyart.map.GateMapper(new_radar, old_radar)
    mapped_radar = gate_mapper.mapped_radar(['reflectivity'])

    # Test point outside of 1 min tolerance
    assert gate_mapper[20, 20] == (None, None)
    assert gate_mapper[4, 4] == (26, 11)
    assert mapped_radar.fields[
        'reflectivity']['data'][26, 11] == old_radar.fields[
                'reflectivity']['data'][4, 4]


def test_gatemapper_gatefilter():
    # Make fake radar with target
    old_radar = pyart.testing.make_target_radar()
    new_radar = deepcopy(old_radar)
    new_radar.latitude["data"] = old_radar.latitude["data"] + 0.001
    new_radar.longitude["data"] = old_radar.longitude["data"] + 0.001
    gatefilter = pyart.filters.GateFilter(old_radar)
    gatefilter.exclude_below('reflectivity', 40)
    gate_mapper = pyart.map.GateMapper(new_radar, old_radar,
                                       gatefilter_src=gatefilter)
    mapped_radar = gate_mapper.mapped_radar(['reflectivity'])
    assert gate_mapper[4, 4] == (26, 11)
    assert np.ma.is_masked(mapped_radar.fields['reflectivity']['data'][26, 11])


import numpy as np


def _steiner_conv_strat(refl, x, y, dx, dy, intense=42, peak_relation=0,
                        area_relation=1, bkg_rad=11000, use_intense=True):
    """
    We perform the Steiner et al. (1995) algorithm for echo classification
    using only the reflectivity field in order to classify each grid point
    as either convective, stratiform or undefined. Grid points are
    classified as follows,

    0 = Undefined
    1 = Stratiform
    2 = Convective
    """
    def convective_radius(ze_bkg, area_relation):
        """
        Given a mean background reflectivity value, we determine via a step
        function what the corresponding convective radius would be.

        Higher background reflectivitives are expected to have larger
        convective influence on surrounding areas, so a larger convective
        radius would be prescribed.
        """
        if area_relation == 0:
            if ze_bkg < 30:
                conv_rad = 1000.
            elif (ze_bkg >= 30) & (ze_bkg < 35.):
                conv_rad = 2000.
            elif (ze_bkg >= 35.) & (ze_bkg < 40.):
                conv_rad = 3000.
            elif (ze_bkg >= 40.) & (ze_bkg < 45.):
                conv_rad = 4000.
            else:
                conv_rad = 5000.

        if area_relation == 1:
            if ze_bkg < 25:
                conv_rad = 1000.
            elif (ze_bkg >= 25) & (ze_bkg < 30.):
                conv_rad = 2000.
            elif (ze_bkg >= 30.) & (ze_bkg < 35.):
                conv_rad = 3000.
            elif (ze_bkg >= 35.) & (ze_bkg < 40.):
                conv_rad = 4000.
            else:
                conv_rad = 5000.

        if area_relation == 2:
            if ze_bkg < 20:
                conv_rad = 1000.
            elif (ze_bkg >= 20) & (ze_bkg < 25.):
                conv_rad = 2000.
            elif (ze_bkg >= 25.) & (ze_bkg < 30.):
                conv_rad = 3000.
            elif (ze_bkg >= 30.) & (ze_bkg < 35.):
                conv_rad = 4000.
            else:
                conv_rad = 5000.

        if area_relation == 3:
            if ze_bkg < 40:
                conv_rad = 0.
            elif (ze_bkg >= 40) & (ze_bkg < 45.):
                conv_rad = 1000.
            elif (ze_bkg >= 45.) & (ze_bkg < 50.):
                conv_rad = 2000.
            elif (ze_bkg >= 50.) & (ze_bkg < 55.):
                conv_rad = 6000.
            else:
                conv_rad = 8000.

        return conv_rad

    def peakedness(ze_bkg, peak_relation):
        """
        Given a background reflectivity value, we determine what the necessary
        peakedness (or difference) has to be between a grid point's
        reflectivity and the background reflectivity in order for that grid
        point to be labeled convective.
        """
        if peak_relation == 0:
            if ze_bkg < 0.:
                peak = 10.
            elif (ze_bkg >= 0.) and (ze_bkg < 42.43):
                peak = 10. - ze_bkg ** 2 / 180.
            else:
                peak = 0.

        elif peak_relation == 1:
            if ze_bkg < 0.:
                peak = 14.
            elif (ze_bkg >= 0.) and (ze_bkg < 42.43):
                peak = 14. - ze_bkg ** 2 / 180.
            else:
                peak = 4.

        return peak

    sclass = np.zeros(refl.shape, dtype=int)
    ny, nx = refl.shape

    for i in range(0, nx):
        # Get stencil of x grid points within the background radius
        imin = np.max(np.array([1, (i - bkg_rad / dx)], dtype=int))
        imax = np.min(np.array([nx, (i + bkg_rad / dx)], dtype=int))

        for j in range(0, ny):
            # First make sure that the current grid point has not already been
            # classified. This can happen when grid points within the
            # convective radius of a previous grid point have also been
            # classified.
            if ~np.isnan(refl[j, i]) & (sclass[j, i] == 0):
                # Get stencil of y grid points within the background radius
                jmin = np.max(np.array([1, (j - bkg_rad / dy)], dtype=int))
                jmax = np.min(np.array([ny, (j + bkg_rad / dy)], dtype=int))

                n = 0
                sum_ze = 0

                # Calculate the mean background reflectivity for the current
                # grid point, which will be used to determine the convective
                # radius and the required peakedness.

                for l in range(imin, imax):
                    for m in range(jmin, jmax):
                        if not np.isnan(refl[m, l]):
                            rad = np.sqrt(
                                (x[l] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                        # The mean background reflectivity will first be
                        # computed in linear units, i.e. mm^6/m^3, then
                        # converted to decibel units.
                            if rad <= bkg_rad:
                                n += 1
                                sum_ze += 10. ** (refl[m, l] / 10.)

                if n == 0:
                    ze_bkg = np.inf
                else:
                    ze_bkg = 10.0 * np.log10(sum_ze / n)

                # Now get the corresponding convective radius knowing the mean
                # background reflectivity.
                conv_rad = convective_radius(ze_bkg, area_relation)

                # Now we want to investigate the points surrounding the current
                # grid point that are within the convective radius, and whether
                # they too are convective, stratiform or undefined.

                # Get stencil of x and y grid points within the convective
                # radius.
                lmin = np.max(
                    np.array([1, int(i - conv_rad / dx)], dtype=int))
                lmax = np.min(
                    np.array([nx, int(i + conv_rad / dx)], dtype=int))
                mmin = np.max(
                    np.array([1, int(j - conv_rad / dy)], dtype=int))
                mmax = np.min(
                    np.array([ny, int(j + conv_rad / dy)], dtype=int))

                if use_intense and (refl[j, i] >= intense):
                    sclass[j, i] = 2

                    for l in range(lmin, lmax):
                        for m in range(mmin, mmax):
                            if not np.isnan(refl[m, l]):
                                rad = np.sqrt(
                                    (x[l] - x[i]) ** 2
                                    + (y[m] - y[j]) ** 2)

                                if rad <= conv_rad:
                                    sclass[m, l] = 2

                else:
                    peak = peakedness(ze_bkg, peak_relation)

                    if refl[j, i] - ze_bkg >= peak:
                        sclass[j, i] = 2

                        for l in range(imin, imax):
                            for m in range(jmin, jmax):
                                if not np.isnan(refl[m, l]):
                                    rad = np.sqrt(
                                        (x[l] - x[i]) ** 2
                                        + (y[m] - y[j]) ** 2)

                                    if rad <= conv_rad:
                                        sclass[m, l] = 2

                    else:
                        # If by now the current grid point has not been
                        # classified as convective by either the intensity
                        # criteria or the peakedness criteria, then it must be
                        # stratiform.
                        sclass[j, i] = 1

    return sclass


def steiner_class_buff(ze, x, y, z, dx, dy, bkg_rad,
                       work_level, intense, peak_relation,
                       area_relation, use_intense):

    zslice = np.argmin(np.abs(z - work_level))
    refl = ze[zslice, :, :]

    area_rel = {"small": 0, "medium": 1, "large": 2, "sgp": 3}
    peak_rel = {"default": 0, "sgp": 1}

    sclass = _steiner_conv_strat(refl, x, y, dx, dy, intense=intense,
                                 peak_relation=peak_rel[peak_relation],
                                 area_relation=area_rel[area_relation],
                                 bkg_rad=11000, use_intense=True)

    return sclass
